from datetime import timedelta

from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    REGISTRATION_DATA_3,
    SUPER_ADMIN_REGISTRATION_DATA,
    WORKFLOW_DATA,
)
from human_lambdas.workflow_handler.tests.util import HLTestCase
from human_lambdas.workflow_handler.utils import STAFF_ORG_ID


class TestListWorkflow(HLTestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id1 = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token1 = response.data["access"]

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA_3)
        self.org_id2 = Organization.objects.get(user__email="bar@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "bar@bar.com", "password": "foowordbar"}
        )
        self.access_token2 = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token1)
        self.wf_name1 = "foowf"
        workflow_data1 = {"name": self.wf_name1, "data": WORKFLOW_DATA["data"]}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id1),
            workflow_data1,
            format="json",
        )

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token2)
        self.wf_name2 = "barwf"
        workflow_data1 = {"name": self.wf_name2, "data": WORKFLOW_DATA["data"]}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id2),
            workflow_data1,
            format="json",
        )

    def test_retrieve_wrong_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token1)

        workflow_obj = Workflow.objects.filter(name=self.wf_name2)
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id1, workflow.pk)
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response.data)

    def test_list_workflow1(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token1)
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id1))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 1)
        result_1 = response.data[0]
        self.assertTrue(result_1.pop("id"))
        self.assertEqual(result_1.pop("name"), self.wf_name1)
        self.assertEqual(result_1["active_users"], [])

    def test_list_workflow2(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token2)
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id2))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 1)
        result_1 = response.data[0]
        self.assertTrue(result_1.pop("id"))
        self.assertEqual(result_1.pop("name"), self.wf_name2)

    def test_when_staff_list_workflow_then_ordered_by_oldest_task(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token2)
        self.admin_token = self.register(
            SUPER_ADMIN_REGISTRATION_DATA, is_super_admin=True
        )["access"]
        self.wf_name2 = "foowf2"
        workflow_data2 = {"name": self.wf_name2, "data": WORKFLOW_DATA["data"]}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id2),
            workflow_data2,
            format="json",
        )

        wfs = list(Workflow.objects.filter(organization__id=self.org_id2).all())
        wfs[0].is_running = True
        wfs[0].save()
        wfs[1].is_running = True
        wfs[1].save()

        t1 = Task(
            workflow=wfs[0],
            inputs={},
            outputs={},
            data={"foo": 2},
        )
        t1.save()
        t1.status = "completed"
        t1.created_at = timezone.now() - timedelta(days=2)
        t1.save()

        t2 = Task(
            workflow=wfs[0],
            inputs={},
            outputs={},
            data={"foo": 2},
        )
        t2.save()

        t3 = Task(
            workflow=wfs[1],
            inputs={},
            outputs={},
            data={"foo": 2},
        )
        t3.save()
        t3.created_at = timezone.now() - timedelta(days=1)
        t3.save()
        t4 = Task(
            workflow=wfs[1],
            inputs={},
            outputs={},
            data={"foo": 2},
        )
        t4.save()

        self.set_user(self.admin_token)
        response = self.client.get("/v1/orgs/{}/workflows".format(STAFF_ORG_ID))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        expected_order = [wfs[1].id, wfs[0].id]
        actual_order = [wf["id"] for wf in list(response.data)]
        assert expected_order == actual_order


class TestListNoWorkflow(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id1 = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

    def test_list_no_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id1))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(response.data, [], response.data)



from rest_framework import status
from rest_framework.test import APITestCase


class TestAPIToken(APITestCase):
    def setUp(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": "foo@bar.com",
                "password": "foowordbar",
                "name": "foo",
                "organization": "barinc",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

    def test_token(self):
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)
        self.assertEqual(self.token, response.data["token"])



from django.core.management.base import BaseCommand

from human_lambdas.workflow_handler.data_mapping import migrate_data
from human_lambdas.workflow_handler.models import Workflow


class Command(BaseCommand):
    help = "migrates inputs and outputs to data"

    def handle(self, *args, **options):
        workflows = Workflow.objects.all()
        for workflow in workflows:
            migrate_data(workflow)
            self.stdout.write(
                self.style.SUCCESS(
                    'Successfully migrated inputs and outputs to data for workflow "%s"'
                    % workflow.id
                )
            )



from typing import Any, Dict

from django.db import models
from rest_hooks.models import AbstractHook
from rest_hooks.signals import hook_event

from human_lambdas.data_handler.data_transformation import transform_int2ext
from human_lambdas.user_handler.models import Notification, Organization, User
from human_lambdas.workflow_handler import regional_storage
from human_lambdas.workflow_handler.fields import JSONField
from human_lambdas.workflow_handler.region import Region

STATUS_MAPPING = {"assigned": "in_progress", "pending": "new"}


class Workflow(models.Model):
    name = models.CharField(max_length=140)
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)
    inputs = JSONField(default=list)
    outputs = JSONField(default=list)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, on_delete=models.CASCADE)
    disabled = models.BooleanField(default=False)
    n_tasks = models.IntegerField(default=0)
    data = JSONField(blank=True, default=list)
    is_running = models.BooleanField(default=False)
    task_description = models.TextField(null=True)
    guidelines_url = models.TextField(null=True)
    pinned_block = models.TextField(null=True)

    def __str__(self):
        return self.name


class Source(models.Model):
    name = models.CharField(max_length=512)
    workflow = models.ForeignKey(Workflow, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, on_delete=models.CASCADE)

    def __str__(self):
        return self.name


class Task(models.Model):
    status = models.CharField(max_length=128, default="new")
    completed_at = models.DateTimeField(null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    assigned_at = models.DateTimeField(null=True)
    assigned_to = models.ForeignKey(User, on_delete=models.CASCADE, null=True)
    session_started_at = models.DateTimeField(null=True)
    handling_time_seconds = models.IntegerField(default=0)
    workflow = models.ForeignKey(Workflow, on_delete=models.CASCADE)
    inputs = JSONField(default=list)
    outputs = JSONField(default=list)
    source = models.ForeignKey(Source, on_delete=models.CASCADE, null=True)
    data = JSONField(blank=True, default=list)
    correct = models.BooleanField(null=True)
    region = models.CharField(max_length=128, null=True)

    def __str__(self):
        return "{0}_task_{1}".format(self.workflow.name, self.pk)

    @classmethod
    def from_db(cls, db, field_names, values):
        task = super(Task, cls).from_db(db, field_names, values)

        if "data" in field_names and task.region:
            region = Region[task.region]
            task.data = regional_storage.retrieve(task.pk, region)

        return task

    def save(
        self,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None,
    ):
        if self.region is None or "data" in self.get_deferred_fields():
            super(Task, self).save()
            return

        region = Region[self.region]

        # Save mutated task in DB *without* region-sensitive data. Beware: this is a hack which protects data sovereignty
        data: Any = self.data
        self.data = {}
        try:
            super(Task, self).save()
            regional_storage.store(self.pk, region, data)
        finally:
            # restore regional data on task
            self.data = data

    def get_status(self):
        return STATUS_MAPPING.get(self.status, self.status)

    def task_completed(self, user):
        hook_event.send(
            sender=self.__class__, action="completed", instance=self, user=user
        )

    def get_updated_status(self) -> Dict[Any, Any]:
        task: Dict[Any, Any] = {
            "id": self.pk,
            "status": self.get_status(),
            "completed_at": self.completed_at,
            "created_at": self.created_at,
            "assigned_at": self.assigned_at,
            "assigned_to": self.assigned_to.name if self.assigned_to else None,
            "queue": self.workflow.pk,
            "data": self.data,
            "source": self.source.pk if self.source else None,
            "n_comments": self.taskactivity_set.filter(action="comment").count(),
        }

        if self.region:
            task["region"] = self.region

        return task

    def get_formatted_task(self, include_data: bool = True) -> Dict[Any, Any]:
        if self.assigned_to:
            worker_id = self.assigned_to.pk
            worker_name = self.assigned_to.name
            worker_email = self.assigned_to.email
        else:
            worker_id, worker_name, worker_email = None, None, None
        if self.source:
            source_name, source_id = self.source.name, self.source.pk
        else:
            source_name, source_id = None, None
        task: Dict[Any, Any] = {
            "id": self.pk,
            "status": self.get_status(),
            "assigned_at": self.assigned_at,
            "assigned_to": worker_id,
            "completed_at": self.completed_at,
            "created_at": self.created_at,
            "completed_by": worker_name,
            "completed_by_email": worker_email,
            "queue": self.workflow.name,
            "queue_id": self.workflow.pk,
            "data": self.data if include_data else None,
            "source": source_name,
            "source_id": source_id,
            "n_comments": self.taskactivity_set.filter(action="comment").count(),
            "correct": self.correct,
            "org_id": self.workflow.organization_id,
        }

        if self.region:
            task["region"] = self.region

        return task

    def get_simple_formatted_task(self):
        if self.assigned_to:
            worker_email = self.assigned_to.email
        else:
            worker_email = None
        if self.source:
            source_name = self.source.name
        else:
            source_name = None
        task: Dict[Any, Any] = {
            "id": self.pk,
            "status": self.get_status(),
            "completed_at": self.completed_at,
            "created_at": self.created_at,
            "completed_by": worker_email,
            "queue": self.workflow.name,
            "queue_id": self.workflow.pk,
            "data": transform_int2ext(self.data),
            "source": source_name,
            "correct": self.correct,
        }

        if self.region:
            task["region"] = self.region

        return task

    def get_formatted_task_external(self):
        serilized_data = self.get_simple_formatted_task()
        return serilized_data  # process_external_completed_tasks(serilized_data)

    def serialize_hook(self, *args, **kwargs):
        return self.get_formatted_task_external()


class WebHook(AbstractHook):
    workflow = models.ForeignKey(Workflow, on_delete=models.CASCADE, default=None)
    is_zapier = models.BooleanField(default=False)


class WorkflowNotification(models.Model):
    workflow = models.ForeignKey(Workflow, on_delete=models.CASCADE)
    enabled = models.BooleanField(default=True)
    notification = models.ForeignKey(Notification, on_delete=models.CASCADE)
    last_notified = models.DateTimeField(null=True)


ACTIVITY_ACTIONS = [
    ("created", "created"),
    ("assigned", "assigned"),
    ("completed", "completed"),
    ("saved", "saved"),
    ("comment", "comment"),
    ("audited_correct", "audited_correct"),
    ("audited_incorrect", "audited_incorrect"),
    ("audited_empty", "audited_empty"),
]


ACTIVITY_SOURCES = [
    ("api", "api"),
    ("zapier", "zapier"),
    ("csv", "csv"),
    ("manual", "manual"),
]


class TaskActivity(models.Model):
    task = models.ForeignKey(Task, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="creator"
    )
    action = models.CharField(max_length=128, choices=ACTIVITY_ACTIONS)
    source = models.CharField(max_length=64, choices=ACTIVITY_SOURCES, null=True)
    filename = models.CharField(max_length=128, null=True)
    assignee = models.ForeignKey(
        User, on_delete=models.CASCADE, null=True, related_name="assignee"
    )
    comment = models.TextField(null=True)



import copy
import logging

from django.db import transaction
from django.db.models import OuterRef, Q, Subquery
from django.shortcuts import get_list_or_404, get_object_or_404
from django.utils import timezone
from drf_yasg2.utils import swagger_auto_schema
from rest_framework import request, serializers
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    RetrieveAPIView,
    RetrieveUpdateAPIView,
)
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin
from rest_framework.parsers import MultiPartParser
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_simplejwt.authentication import JWTAuthentication

from human_lambdas.data_handler.csv_utils import process_csv
from human_lambdas.data_handler.data_sync import sync_workflow_task
from human_lambdas.external.authentication import TokenAuthentication
from human_lambdas.user_handler.permissions import IsAuthorized, IsOrgAdmin
from human_lambdas.workflow_handler.utils import is_force

from .models import Source, Task, TaskActivity, User, WebHook, Workflow
from .serializers import (
    HookSerializer,
    PendingTaskSerializer,
    TaskSerializer,
    WorkflowSerializer,
)
from .utils import STAFF_ORG_ID, TaskPagination, decode_csv, notify_slack


class RUWebhookView(RetrieveUpdateAPIView, CreateModelMixin):
    permission_classes = (
        IsAuthenticated,
        IsOrgAdmin,
    )
    authentication_classes = (TokenAuthentication, JWTAuthentication)
    serializer_class = HookSerializer

    def get_queryset(self):
        return WebHook.objects.filter(workflow__pk=self.kwargs["workflow_id"])

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        return queryset.first()

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        if not instance and request.data["target"]:
            response = self.create(request)
        elif not instance and not request.data["target"]:
            response = Response({"status_code": 204}, status=204)
        elif not request.data["target"]:
            response = self.destroy(request)
        else:
            self.check_object_permissions(self.request, instance)
            serializer = self.get_serializer(
                instance, data=request.data, partial=partial
            )
            serializer.is_valid(raise_exception=True)
            self.perform_update(serializer)
            if getattr(instance, "_prefetched_objects_cache", None):
                # If 'prefetch_related' has been applied to a queryset, we need to
                # forcibly invalidate the prefetch cache on the instance.
                instance._prefetched_objects_cache = {}
            response = Response(serializer.data)
        return response

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.check_object_permissions(self.request, instance)
        self.perform_destroy(instance)
        return Response(status=204)

    def perform_destroy(self, instance):
        instance.delete()


logger = logging.getLogger(__name__)


class CreateWorkflowView(CreateAPIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)
    serializer_class = WorkflowSerializer

    def get_queryset(self):

        return Workflow.objects.filter(
            Q(organization__pk=self.kwargs["org_id"]) & Q(disabled=False)
        )

    def perform_create(self, serializer):
        queryset = self.get_queryset()
        if self.request.data["name"] in queryset.values_list("name", flat=True):
            raise serializers.ValidationError("Workflow with same name already exists")
        serializer.save()


class ListWorkflowView(ListAPIView):
    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = WorkflowSerializer
    authentication_classes = (TokenAuthentication, JWTAuthentication)

    def get_queryset(self):
        queryset = Workflow.objects.filter(Q(disabled=False))

        if self.kwargs["org_id"] == STAFF_ORG_ID:
            task_age_subq = Task.objects.filter(
                Q(workflow__id=OuterRef("pk")) & ~Q(status="completed")
            ).order_by("created_at")
            queryset = queryset.annotate(
                oldest_created=Subquery(task_age_subq.values("created_at")[:1])
            ).order_by("oldest_created")
            queryset = queryset.filter(Q(is_running=True))
        else:
            queryset = queryset.filter(Q(organization__pk=self.kwargs["org_id"]))

        task_status = self.request.query_params.get("task_status")
        if task_status:
            queryset = queryset.filter(task__status=task_status).distinct()
        return queryset

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        serializer = self.get_serializer(queryset, many=True)

        for d in serializer.data:
            if self.kwargs["org_id"] != STAFF_ORG_ID:
                del d["org_id"]
        return Response(serializer.data)


class BaseWorkflowView(RetrieveAPIView):
    """
    Retrieve and Update for now, will add delete here later
    """

    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = WorkflowSerializer
    authentication_classes = (TokenAuthentication, JWTAuthentication)

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs["context"] = self.get_serializer_context(*args, **kwargs)
        return serializer_class(*args, **kwargs)

    def get_serializer_context(self, *args, **kwargs):
        """
        Extra context provided to the serializer class.
        """
        context = {
            "request": self.request,
            "format": self.format_kwarg,
            "view": self,
        }
        try:
            if not kwargs["data"]["webhook"]:
                context["remove_webhook"] = not kwargs["data"].pop("webhook")
        except KeyError:
            pass
        return context

    def get_queryset(self):
        if self.kwargs["org_id"] == STAFF_ORG_ID:
            queryset = Q(disabled=False) & Q(pk=self.kwargs["workflow_id"])
        else:
            queryset = (
                Q(organization__pk=self.kwargs["org_id"])
                & Q(disabled=False)
                & Q(pk=self.kwargs["workflow_id"])
            )

        return Workflow.objects.filter(queryset)

    def get_object(self):
        obj = get_object_or_404(self.get_queryset(), id=self.kwargs["workflow_id"])
        return obj

    def _retrieve(self):
        obj = get_object_or_404(self.get_queryset())
        workflow = self.serializer_class(obj).data
        if (
            hasattr(obj, "webhook_set")
            and obj.webhook_set.filter(is_zapier=False).exists()
        ):
            workflow["webhook"] = {
                "target": WebHook.objects.get(workflow=obj, is_zapier=False).target
            }

        return workflow

    def retrieve(self, request, *args, **kwargs):
        return Response(self._retrieve())


class InternalWorkflowView(UpdateModelMixin, BaseWorkflowView):
    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response(serializer.data)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class ExternalWorkflowView(BaseWorkflowView):
    def retrieve(self, request, *args, **kwargs):
        workflow = self._retrieve()
        for block in workflow["data"]:
            block.pop("layout", None)
            block.pop("_id", None)

        return Response(workflow)


class FileUploadViewQuerySerializer(serializers.Serializer):
    region = serializers.CharField(required=False)


class FileUploadView(APIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)
    parser_classes = [MultiPartParser]

    def get_queryset(self):

        workflows = Workflow.objects.filter(
            Q(organization__pk=self.kwargs["org_id"]) & Q(disabled=False)
        )
        return workflows.filter(pk=self.kwargs["workflow_id"])

    @swagger_auto_schema(query_serializer=FileUploadViewQuerySerializer)
    def post(self, request: request.Request, *args, **kwargs):
        file_obj = request.data["file"]
        workflow: Workflow = get_object_or_404(self.get_queryset())
        content = decode_csv(file_obj)
        filename = request.data["file"].name
        source = Source(name=filename, workflow=workflow, created_by=request.user)
        source.save()
        try:
            process_csv(
                content,
                workflow=workflow,
                source=source,
                user=request.user,
                filename=filename,
                region=request.query_params.get("region", None),
            )
        except Exception as exception:
            return Response(
                {"status_code": 400, "errors": [{"message": str(exception)}]},
                status=400,
            )
        if workflow.is_running:
            notify_slack(f"bulk upload for {workflow.name}", request)

        return Response(
            {
                "status_code": 200,
                "message": f"File {filename} was processed and task created",
            },
            status=200,
        )


class ListTaskView(ListAPIView):
    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = TaskSerializer

    def get_queryset(self):

        workflows = Workflow.objects.filter(
            Q(disabled=False) & Q(organization__pk=self.kwargs["org_id"])
        )
        return (
            Task.objects.defer("data")
            .filter(Q(workflow__in=workflows) & Q(workflow=self.kwargs["workflow_id"]))
            .order_by("-created_at")
        )

    def list(self, request, *args, **kwargs):
        workflow = Workflow.objects.get(id=kwargs["workflow_id"])
        obj = get_list_or_404(self.get_queryset(), workflow=workflow)
        serializer = self.serializer_class(obj, many=True)
        return Response(serializer.data)


class ListNonCompleteTaskView(ListTaskView):
    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = PendingTaskSerializer
    pagination_class = TaskPagination

    def get_queryset(self, *args, **kwargs):

        workflow = Workflow.objects.filter(
            Q(pk=self.kwargs["workflow_id"]) & Q(disabled=False)
        )
        return (
            Task.objects.defer("data")
            .filter(Q(workflow=workflow.first()) & ~Q(status="completed"))
            .order_by("created_at")
        )

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        filtered_queryset = self.filter_queryset(queryset)
        page = self.paginate_queryset(filtered_queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)
        serializer = self.serializer_class(queryset.all(), many=True)
        return Response(serializer.data)


class RUDTaskView(RetrieveUpdateAPIView):
    """
    Retrieve and Update for now, will add delete here later
    """

    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = TaskSerializer

    def get_queryset(self):

        if self.kwargs["org_id"] == STAFF_ORG_ID:
            queryset = Q(disabled=False)
        else:
            queryset = Q(organization__pk=self.kwargs["org_id"]) & Q(disabled=False)

        workflows = Workflow.objects.filter(queryset)
        return Task.objects.defer("data").filter(
            Q(workflow__in=workflows) & Q(workflow=self.kwargs["workflow_id"])
        )

    def get_object(self):
        workflow = Workflow.objects.get(id=self.kwargs["workflow_id"])
        queryset = self.get_queryset()
        obj = get_object_or_404(queryset, id=self.kwargs["task_id"], workflow=workflow)
        return obj

    def retrieve(self, request, *args, **kwargs):
        workflow = Workflow.objects.get(id=self.kwargs["workflow_id"])
        queryset = self.get_queryset()
        obj = get_object_or_404(queryset, id=self.kwargs["task_id"], workflow=workflow)
        if obj.status in ["new", "pending", "open"]:
            obj.assigned_to = request.user
            obj.status = "in_progress"
            obj.assigned_at = timezone.now()
            obj.save()
            TaskActivity(
                task=obj,
                created_by=request.user,
                action="assigned",
                assignee=request.user,
            ).save()
        sync_workflow_task(workflow, obj)
        task = self.serializer_class(obj).data
        task["status_code"] = 200
        return Response(task, status=200)

    def perform_update(self, serializer, *args, **kwargs):
        serializer.save(
            owner=self.request.user,
            submit_task=True,
            force=is_force(self.request.query_params),
        )


class RefreshTaskView(RUDTaskView):
    def retrieve(self, request, *args, **kwargs):
        workflow = Workflow.objects.get(id=self.kwargs["workflow_id"])
        queryset = self.get_queryset()
        obj = get_object_or_404(queryset, id=self.kwargs["task_id"], workflow=workflow)
        sync_workflow_task(workflow, obj)
        task = self.serializer_class(obj).data
        task["status_code"] = 200
        return Response(task, status=200)


class SaveTaskView(RUDTaskView):
    def perform_update(self, serializer, *args, **kwargs):
        serializer.save(
            owner=self.request.user,
            submit_task=False,
            force=is_force(self.request.query_params),
        )


class NextTaskView(APIView):
    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = TaskSerializer

    def get_queryset(self):

        if self.kwargs["org_id"] == STAFF_ORG_ID:
            queryset = Q(disabled=False)
        else:
            queryset = Q(organization__pk=self.kwargs["org_id"]) & Q(disabled=False)

        workflows = Workflow.objects.filter(queryset)

        return (
            Task.objects.defer("data")
            .filter(Q(workflow__in=workflows) & Q(workflow=self.kwargs["workflow_id"]))
            .order_by("created_at")
        )

    def get(self, request, *args, **kwargs):
        workflow = Workflow.objects.get(id=kwargs["workflow_id"])
        queryset = self.get_queryset()

        # 1 get assigned to self
        with transaction.atomic():
            obj = (
                queryset.filter(Q(status="assigned") | Q(status="in_progress"))
                .filter(assigned_to=request.user)
                .first()
            )
            if obj:
                obj.assigned_at = timezone.now()
                obj.save()
                sync_workflow_task(workflow, obj)
                task = self.serializer_class(obj).data
                task["status_code"] = 200
                return Response(task, status=200)

        # 2 get first pending, can take +1.3s
        with transaction.atomic():
            obj = (
                queryset.select_for_update()
                .filter(Q(status="pending") | Q(status="new"))
                .first()  # TODO: Remove pending in the future
            )
            if not obj:
                obj = queryset.select_for_update().filter(status="open").first()
            if obj:
                obj.status = "in_progress"
                obj.assigned_to = request.user
                obj.assigned_at = timezone.now()
                obj.save()
                TaskActivity(
                    task=obj,
                    created_by=request.user,
                    action="assigned",
                    assignee=request.user,
                ).save()
                sync_workflow_task(workflow, obj)
                task = self.serializer_class(obj).data
                task["status_code"] = 200
                return Response(task, status=200)
        return Response(status=204)


class AssignTaskView(APIView):
    permission_classes = (IsAuthenticated, IsAuthorized)

    def get_queryset(self):

        if self.kwargs["org_id"] == STAFF_ORG_ID:
            queryset = Q(disabled=False)
        else:
            queryset = Q(organization__pk=self.kwargs["org_id"]) & Q(disabled=False)

        workflows = Workflow.objects.filter(queryset)

        return Task.objects.defer("data").filter(
            Q(workflow__in=workflows)
            & Q(workflow=self.kwargs["workflow_id"])
            & Q(pk=self.kwargs["task_id"])
        )

    def get_userset(self):
        return User.objects.filter(
            Q(organization__pk=self.request.user.current_organization_id)
        )

    def post(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        task = get_object_or_404(queryset)
        if task.status == "completed":
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": f"Task {task.pk} is already completed!"}],
                },
                status=400,
            )
        assigned_to = request.data["assigned_to"]
        if not assigned_to:
            task.assigned_to = None
            task.status = "open"
            task.assigned_at = None
            task.save()
            TaskActivity(
                task=task, created_by=request.user, action="assigned", assignee=None
            ).save()
            return Response(
                {
                    "status_code": 200,
                    "message": f"Task {task.pk} was unassigned successfully!",
                    "assigned_to": None,
                },
                status=200,
            )
        else:
            users = self.get_userset()
            user = get_object_or_404(users, pk=assigned_to)
            task.assigned_to = user
            task.status = "in_progress"
            task.assigned_at = timezone.now()
            task.save()
            TaskActivity(
                task=task, created_by=request.user, action="assigned", assignee=user
            ).save()
            return Response(
                {
                    "status_code": 200,
                    "message": f"Task {task.pk} was assigned successfully"
                    f" to {user.name}!",
                    "assigned_to": task.pk,
                },
                status=200,
            )


class CreateTaskFormView(CreateAPIView):
    permission_classes = (IsAuthenticated, IsAuthorized)
    serializer_class = TaskSerializer

    def get_queryset(self):
        workflows = self.workflow_queryset()
        return Task.objects.defer("data").filter(
            Q(workflow__in=workflows) & Q(workflow=self.kwargs["workflow_id"])
        )

    def workflow_queryset(self):

        workflows = Workflow.objects.filter(
            Q(organization__pk=self.kwargs["org_id"]) & Q(disabled=False)
        )
        return workflows

    def get(self, request, *args, **kwargs):
        workflow = get_object_or_404(
            self.workflow_queryset(), pk=self.kwargs["workflow_id"]
        )
        result = []
        for idata in workflow.data:
            if idata[idata["type"]].get("read_only") or idata["type"] in [
                "video",
                "image",
                "audio",
                "embed",
                "named_entity_recognition",
                "pdf",
                "bounding_boxes",
            ]:
                read_data = copy.deepcopy(idata)
                if "layout" in read_data:
                    del read_data["layout"]
                result.append(read_data)
        return Response({"status_code": 200, "data": result}, status=200)

    def post(self, request, *args, **kwargs):
        workflow: Workflow = get_object_or_404(
            self.workflow_queryset(), pk=self.kwargs["workflow_id"]
        )
        ids = [i["id"] for i in request.data["data"]]
        for idata in workflow.data:
            if idata["id"] not in ids:
                request.data["data"].append(idata)

        if workflow.is_running:
            notify_slack(f"Task created via form for {workflow.name}", request)

        return self.create(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save(source_name="manual")



import datetime
import logging

from django.utils import timezone
from django.utils.timezone import make_aware
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Invitation, Organization, User

logger = logging.getLogger(__name__)


class TestInvite(APITestCase):
    def setUp(self):
        self.preset_user_name = "foo"
        self.preset_user_email = "foo@bar.com"
        self.preset_changed_email = "bar@foo.com"
        self.organization_name = "fooinc"
        self.preset_user_password = "foowordbar"

        user = User(name=self.preset_user_name, email=self.preset_user_email)
        user.set_password(self.preset_user_password)
        user.save()
        self.user = user

        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

        organization = Organization(name=self.organization_name)
        organization.save()
        self.organization = organization
        self.org_id = organization.id
        organization.add_admin(user)

        self.existing_recipient, time = (
            "lambda@sigma.com",
            str(datetime.datetime.now()),
        )
        self.token = hash(str(self.existing_recipient) + str(self.org_id) + str(time))
        aware_expiry_date = make_aware(datetime.datetime.now() + datetime.timedelta(30))

        existing_invite = Invitation(
            email=self.preset_user_email,
            organization=self.organization,
            invited_by=self.user,
            token=self.token,
            expires_at=aware_expiry_date,
            invite_link="llll",
        )
        existing_invite.save()

    def test_invitation_get_call_no_invite(self):
        response = self.client.get(
            "/v1/users/invitation/vv3w87nvw3703yvw07vhs7vsnhs0vv4t7ehom"
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_invitation_get_call(self):
        response = self.client.get("/v1/users/invitation/{0}".format(self.token))
        self.assertEqual(response.data["invitation_email"], self.preset_user_email)
        self.assertEqual(response.data["invitation_org"], str(self.organization))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_invitation_post_call_new_user(self):
        recipient = "new@user.com"
        _ = self.client.post(
            "/v1/users/invite",
            {"emails": recipient, "organization_id": self.org_id},
        )

        token = hash(str(recipient) + str(self.org_id) + str(datetime.datetime.now()))

        new_user_invite = Invitation(
            email=recipient,
            organization=self.organization,
            invited_by=self.user,
            token=token,
            expires_at=make_aware(datetime.datetime.now() + datetime.timedelta(30)),
        )
        new_user_invite.save()

        response = self.client.post(
            "/v1/users/invitation/{0}".format(token),
            {"name": "sean", "password": "foowordbar"},
        )
        self.assertEqual(response.data["message"], "Your account has been created!")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_invitation_post_call_existing_user(self):
        new_user = User(email="new@user.com")
        new_user.save()
        _ = self.client.post(
            "/v1/users/invite",
            {"emails": "new@user.com", "organization_id": self.org_id},
        )
        token = "s8ni3fisme03msi0s3emimc"
        new_invite = Invitation(
            email="new@user.com",
            organization=self.organization,
            invited_by=self.user,
            token=token,
            expires_at=make_aware(datetime.datetime.now() + datetime.timedelta(30)),
        )
        new_invite.save()
        response = self.client.post(
            "/v1/users/invitation/{0}".format(token),
            {"name": "sean", "password": "foowordbar"},
        )
        self.assertEqual(response.data["message"], "Success!")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_invitation_expiration_30_days(self):
        _ = self.client.post(
            "/v1/users/invite",
            {"emails": self.existing_recipient, "organization_id": self.org_id},
        )

        token = hash(
            str(self.existing_recipient)
            + str(self.org_id)
            + str(datetime.datetime.now())
        )

        expired_invite = Invitation(
            email=self.existing_recipient,
            organization=self.organization,
            invited_by=self.user,
            token=token,
            expires_at=make_aware(datetime.datetime.now() - datetime.timedelta(1)),
        )
        expired_invite.save()

        response = self.client.post(
            "/v1/users/invitation/{0}".format(token),
            {"name": "sean", "password": "foowordbar"},
        )
        self.assertEqual(
            response.data["errors"][0]["message"], "this token has expired!"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_invitation_post_call_already_joined_org(self):
        _ = self.client.post(
            "/v1/users/invite",
            {"emails": self.existing_recipient, "organization_id": self.org_id},
        )
        response = self.client.post(
            "/v1/users/invitation/{0}".format(self.token),
            {"name": "sean", "password": "foowordbar"},
        )
        self.assertEqual(
            response.data["errors"][0]["message"],
            "this organization has already been joined",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_invitation_post_deletes_invites_after_new_user(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        recipient = "new@user.com"
        _ = self.client.post(
            "/v1/users/invite",
            {"emails": recipient, "organization_id": self.org_id},
        )

        token = hash(str(recipient) + str(self.org_id) + str(datetime.datetime.now()))

        new_user_invite = Invitation(
            email=recipient,
            organization=self.organization,
            invited_by=self.user,
            token=token,
            expires_at=make_aware(datetime.datetime.now() + datetime.timedelta(30)),
            invite_link="lkjhlkjh",
        )
        new_user_invite.save()

        response = self.client.post(
            "/v1/users/invitation/{0}".format(token),
            {"name": "sean", "password": "foowordbar"},
        )
        self.assertEqual(response.data["message"], "Your account has been created!")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.get("/v1/orgs/{0}/invite".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        for invite in response.data["invited_users"]:
            self.assertNotEqual(invite["email"], recipient)

    def test_invitation_post_deletes_invites_after_existing_user(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        new_user = User(email="new@user.com")
        new_user.save()
        _ = self.client.post(
            "/v1/users/invite",
            {"emails": "new@user.com", "organization_id": self.org_id},
        )
        token = "s8ni3fisme03msi0s3emimc"
        new_invite = Invitation(
            email="new@user.com",
            organization=self.organization,
            invited_by=self.user,
            token=token,
            expires_at=make_aware(datetime.datetime.now() + datetime.timedelta(30)),
        )
        new_invite.save()
        response = self.client.post(
            "/v1/users/invitation/{0}".format(token),
            {"name": "sean", "password": "foowordbar"},
        )
        self.assertEqual(response.data["message"], "Success!")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        response = self.client.get("/v1/orgs/{0}/invite".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        for invite in response.data["invited_users"]:
            self.assertNotEqual(invite["email"], "new@user.com")

    def test_when_duplicate_invite_then_prev_invite_overwritten(self):
        Invitation.objects.get_queryset().delete()
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        other_org = Organization(name="baz")
        other_org.save()
        email = "foo2@bar.com"
        other_email = "other@bar.com"
        test_invs = (
            (email, self.organization),
            (other_email, self.organization),
            (email, other_org),
        )
        initial_expiry = timezone.now() + timezone.timedelta(days=1)

        for (email, org) in test_invs:
            Invitation(
                email=email,
                organization=org,
                invited_by=self.user,
                token="asdf",
                expires_at=initial_expiry,
                admin=False,
            ).save()

        # act
        response = self.client.post(
            f"/v1/orgs/{self.org_id}/invite",
            {"emails": email, "organization_id": self.org_id},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        updated_invite = Invitation.objects.get(
            email=email, organization=self.organization
        )
        self.assertGreater(updated_invite.expires_at, initial_expiry)
        unaffected_invite1 = Invitation.objects.get(
            email=other_email, organization=self.organization
        )
        unaffected_invite2 = Invitation.objects.get(email=email, organization=other_org)
        self.assertEqual(
            unaffected_invite1.expires_at, unaffected_invite2.expires_at, initial_expiry
        )



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.user_handler.notifications import send_notification
from human_lambdas.workflow_handler.models import (
    Workflow,
    WorkflowNotification,
)
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    WORKFLOW_DATA,
)


class TestNotifications(APITestCase):
    def setUp(self):

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow = Workflow.objects.get(pk=response.data["id"])
        self.workflow_id = workflow.pk

    def test_notification_retrieval(self):
        response = self.client.get(
            "/v1/users/notifications",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["queue_notifications"]), 1)
        self.assertTrue(response.data["queue_notifications"][0]["enabled"])

    def test_notification_update(self):
        response = self.client.patch("/v1/users/notifications", data={"enabled": False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertFalse(response.data["enabled"])

    def test_notification_update_workflow(self):
        response = self.client.patch(
            "/v1/users/notifications",
            data={
                "queue_notifications": [
                    {"queue_id": self.workflow_id, "enabled": False}
                ]
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertFalse(response.data["queue_notifications"][0]["enabled"])

    def test_send_notification(self):
        workflow = Workflow.objects.get(pk=self.workflow_id)
        wfnotifications = WorkflowNotification.objects.filter(workflow=workflow).all()
        for wfn in wfnotifications:
            self.assertIsNone(wfn.last_notified)
        send_notification(workflow=workflow)
        wfnotifications = WorkflowNotification.objects.filter(workflow=workflow).all()
        for wfn in wfnotifications:
            self.assertIsNotNone(wfn.last_notified)



import os

import cchardet
from rest_framework.test import APITestCase

from human_lambdas.workflow_handler.utils import unidecode

from . import DATA_PATH


class TestEncodings(APITestCase):
    def setUp(self):
        self.raw_text = []
        self.expected = []
        for efile in os.listdir(os.path.join(DATA_PATH, "encodings")):
            with open(os.path.join(DATA_PATH, "encodings", efile), "rb") as f:
                raw_text = f.read()
                self.raw_text.append(raw_text)
                self.expected.append(
                    raw_text.decode(cchardet.detect(raw_text)["encoding"])
                )

    def test_ecodings(self):
        for raw_text, expected in zip(self.raw_text, self.expected):
            self.assertEqual(unidecode(raw_text), expected)



import copy

"""
Transformations handle internal<>external conversions.

external->internal is most important because we want to avoid garbage entering the system.
Still, there's a validation stage after this, so we don't need to validate here.
The goal is to pick up any relevant data that's meaningful for the type in question
"""


def list_int2ext(int_data):
    return [
        ivalue["value"] if isinstance(ivalue, dict) else ivalue
        for ivalue in int_data["value"]
    ]


def form_int2ext(int_data):
    data = int_data.get("data")
    return {idata["id"]: idata[idata["type"]]["value"] for idata in data}


def ner_int2ext(int_data):
    int_value = default_int2ext(int_data)
    entities = []
    for entity in int_data["entities"]:
        entity["category"] = entity["tag"]
        del entity["tag"]
        if "text" in entity:
            del entity["text"]
        if "color" in entity:
            del entity["color"]
        entities.append(entity)
    return {"text": int_value, "entities": entities}


def bounding_boxes_int2ext(int_data):
    # Parse 'value'
    int_value = default_int2ext(int_data)

    # Guarantee 'objects' exists
    int_value["objects"] = int_value.get("objects", [])

    # Convert percent to decimal percentage representation
    for i, obj in enumerate(int_value["objects"]):
        obj["x"] = round(obj["x"] / 100, 4)
        obj["y"] = round(obj["y"] / 100, 4)
        obj["w"] = round(obj["w"] / 100, 4)
        obj["h"] = round(obj["h"] / 100, 4)
        int_value["objects"][i] = obj
    return int_value


def default_int2ext(int_data):
    return int_data.get("value")


TRANSFORM_INT2EXT_STATES = {
    "list": list_int2ext,
    "form_sequence": form_int2ext,
    "named_entity_recognition": ner_int2ext,
    "bounding_boxes": bounding_boxes_int2ext,
}


def transform_int2ext(data):  # rename: transform_int2ext
    data = {
        idata["id"]: TRANSFORM_INT2EXT_STATES.get(idata["type"], default_int2ext)(
            idata.get(idata["type"], {})
        )
        for idata in data
    }
    return data


# Transforming external to internal
def default_ext2int(task_data, request_data):
    task_data[task_data["type"]]["value"] = request_data[task_data["id"]]


def list_ext2int(task_data, request_data):
    if task_data[task_data["type"]]["subtype"] == "number":
        input_value = [float(i) for i in request_data[task_data["id"]]]
    else:
        input_value = request_data[task_data["id"]]
    task_data[task_data["type"]]["value"] = input_value


def form_ext2int(task_data, request_data):
    for fs_data in task_data[task_data["type"]]["data"]:
        fs_data[fs_data["type"]]["value"] = request_data[task_data["id"]].get(
            fs_data["id"], None
        )


def ner_ext2int(task_data, request_data):
    # Special case: accept string as its value
    if isinstance(request_data[task_data["id"]], str):
        task_data[task_data["type"]]["value"] = request_data[task_data["id"]]
        task_data[task_data["type"]]["entities"] = []
        return
    # Normal case: standard NER schema
    if (
        not isinstance(request_data[task_data["id"]], dict)
        or "text" not in request_data[task_data["id"]]
    ):
        return
    task_data[task_data["type"]]["value"] = request_data[task_data["id"]]["text"]
    entities = []
    for entity in request_data[task_data["id"]].get("entities", []):
        if "category" in entity:
            entity["tag"] = entity["category"]
            del entity["category"]
        # if it doesn't have 'category' or 'start' or 'end' data validation will fail later on
        entities.append(entity)
    task_data[task_data["type"]]["entities"] = entities


def bounding_boxes_ext2int(task_data, request_data):
    # Try to extract data if value has the right format
    if isinstance(request_data[task_data["id"]], dict):
        task_data[task_data["type"]]["value"] = {}
        task_data[task_data["type"]]["value"]["image"] = request_data[
            task_data["id"]
        ].get("image")
        task_data[task_data["type"]]["value"]["objects"] = request_data[
            task_data["id"]
        ].get("objects", [])
        for i in range(len(task_data[task_data["type"]]["value"]["objects"])):
            obj = task_data[task_data["type"]]["value"]["objects"][i]
            obj["x"] = (
                round(obj["x"] * 100, 2)
                if isinstance(obj.get("x"), (int, float))
                else obj.get("x")
            )
            obj["y"] = (
                round(obj["y"] * 100, 2)
                if isinstance(obj.get("y"), (int, float))
                else obj.get("y")
            )
            obj["w"] = (
                round(obj["w"] * 100, 2)
                if isinstance(obj.get("w"), (int, float))
                else obj.get("w")
            )
            obj["h"] = (
                round(obj["h"] * 100, 2)
                if isinstance(obj.get("h"), (int, float))
                else obj.get("h")
            )
            task_data[task_data["type"]]["value"]["objects"][i] = obj


TRANSFORM_EXT2INT_STATES = {
    "list": list_ext2int,
    "form_sequence": form_ext2int,
    "named_entity_recognition": ner_ext2int,
    "bounding_boxes": bounding_boxes_ext2int,
}


def transform_ext2int(workflow_data, request_data):
    formatted_data = []
    for w_data in workflow_data:
        task_data = copy.deepcopy(w_data)
        # remove layout since it should not be included in tasks
        if "layout" in task_data:
            del task_data["layout"]

        # if there are any cases left which do not include the type data
        if isinstance(task_data.get("type"), dict):
            task_data[task_data["type"]] = {}
        if w_data["id"] in request_data:
            TRANSFORM_EXT2INT_STATES.get(w_data["type"], default_ext2int)(
                task_data, request_data
            )
        else:
            task_data[task_data["type"]]["value"] = None
        formatted_data.append(task_data)
    return formatted_data



from human_lambdas.workflow_handler.serializers import TaskMetadataSerializer


class ExternalTaskSerializer(TaskMetadataSerializer):
    def to_representation(self, instance):
        return instance.get_formatted_task_external()



import copy

from .data_validation import data_validation

"""
Every time a task is fetched, we have to merge the task's data values with the workflow object.
We store these separately because a workflow's blocks configuration can change.
If it does, we want to reflect these changes on any task that is not completed.

The entry function is sync_workflow_task and should only be called by
those endpoints which are meant to return a non-complete task and
such task should not be persisted to the DB post-sync.
"""


def iterate_matching(workflow_data, task_data):
    for wdata in workflow_data:
        task_exists = False
        for tdata in task_data:
            if wdata["id"] == tdata["id"]:
                task_exists = True
                yield wdata, tdata
        if not task_exists:
            yield wdata, None


def form_sync(workflow_data, task_data):
    if task_data is None:
        return
    for w_data, t_data in iterate_matching(
        workflow_data[workflow_data["type"]]["data"],
        task_data[task_data["type"]]["data"],
    ):  # TODO: valdation of data here
        w_data[w_data["type"]]["value"] = t_data[t_data["type"]].get("value")
    if "history" in task_data[task_data["type"]]:
        history = task_data[task_data["type"]]["history"]
    else:
        history = []
    workflow_data[workflow_data["type"]]["history"] = history


def ner_sync(workflow_data, task_data):
    # Value gets cloned if it exists, otherwise we don't set it at all
    # Setting value to null breaks the frontend
    if task_data is not None and "value" in task_data[workflow_data["type"]]:
        workflow_data[workflow_data["type"]]["value"] = task_data[
            task_data["type"]
        ].get("value")

    # If the user wants to use the placeholder and value is null
    if (
        workflow_data[workflow_data["type"]].get("value") is None
        and workflow_data[workflow_data["type"]].get("use_placeholder", False) is True
    ):
        workflow_data[workflow_data["type"]]["value"] = workflow_data[
            workflow_data["type"]
        ].get("placeholder")

    # Entities always gets cloned
    workflow_data[workflow_data["type"]]["entities"] = (
        task_data[task_data["type"]].get("entities", [])
        if task_data is not None
        else []
    )


def bb_sync(workflow_data, task_data):
    # Check task_data has the right format, and set it to the right empty state if not
    if (
        isinstance(task_data, dict) is False
        or isinstance(task_data.get(workflow_data["type"]), dict) is False
        or isinstance(task_data[workflow_data["type"]].get("value"), dict) is False
    ):
        # task_data has no useful data, assume it's empty
        # Initialize value object
        workflow_data[workflow_data["type"]]["value"] = {}

        # Intiialize image to null or placeholder
        workflow_data[workflow_data["type"]]["value"]["image"] = (
            workflow_data[workflow_data["type"]]["placeholder"]
            if workflow_data[workflow_data["type"]].get("use_placeholder", False)
            else None
        )

        # Initialize object list
        workflow_data[workflow_data["type"]]["value"]["objects"] = []

    else:
        # Initialize "value" - wf should never have value
        workflow_data[workflow_data["type"]]["value"] = {}

        # Set image to value or fallback to placeholder or null
        workflow_data[workflow_data["type"]]["value"]["image"] = (
            task_data[workflow_data["type"]]["value"]["image"]
            if task_data[workflow_data["type"]]["value"].get("image") is not None
            else (
                workflow_data[workflow_data["type"]]["placeholder"]
                if workflow_data[workflow_data["type"]].get("use_placeholder", False)
                else None
            )
        )

        # Set objects
        workflow_data[workflow_data["type"]]["value"]["objects"] = task_data[
            workflow_data["type"]
        ]["value"].get("objects", [])


def default_data_sync(workflow_data, task_data):
    workflow_data[workflow_data["type"]]["value"] = (
        task_data[task_data["type"]].get("value", None)
        if task_data is not None
        else None
    )
    if (
        workflow_data[workflow_data["type"]]["value"] is None
        and workflow_data[workflow_data["type"]].get("use_placeholder", False) is True
    ):
        workflow_data[workflow_data["type"]]["value"] = workflow_data[
            workflow_data["type"]
        ].get("placeholder")


SYNC_STATES = {
    "form_sequence": form_sync,
    "named_entity_recognition": ner_sync,
    "bounding_boxes": bb_sync,
}


def sync_workflow_task(workflow, task):
    # retrieve(task)  # move to new fn which also fetches task?
    if task.status != "completed":
        updated_data = []
        for workflow_data, task_data in iterate_matching(workflow.data, task.data):
            final_data = copy.deepcopy(workflow_data)
            SYNC_STATES.get(final_data["type"], default_data_sync)(
                final_data, task_data
            )
            updated_data.append(final_data)
        task.data = data_validation(updated_data)



# Generated by Django 2.2.5 on 2019-12-19 11:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0004_auto_20191216_1841"),
    ]

    operations = [
        migrations.RenameField(
            model_name="task",
            old_name="completed",
            new_name="completed_at",
        ),
        migrations.RemoveField(
            model_name="task",
            name="name",
        ),
        migrations.AddField(
            model_name="task",
            name="status",
            field=models.CharField(default="incomplete", max_length=128),
        ),
    ]



from django.contrib import admin

from .models import (
    ForgottenPassword,
    Invitation,
    Notification,
    Organization,
    User,
)

# Register your models here.

admin.site.register(User)
admin.site.register(Organization)
admin.site.register(Invitation)
admin.site.register(ForgottenPassword)
admin.site.register(Notification)



from django.contrib.auth.models import AbstractBaseUser, BaseUserManager
from django.db import models


class Notification(models.Model):
    enabled = models.BooleanField(default=True)

    def format_output(self):
        return {
            "enabled": self.enabled,
            "queue_notifications": [
                {
                    "queue_id": wnotification.workflow.pk,
                    "queue_name": wnotification.workflow.name,
                    "enabled": wnotification.enabled,
                }
                for wnotification in self.workflownotification_set.filter(
                    workflow__disabled=False
                ).all()
            ],
        }


class LowercaseEmailField(models.EmailField):
    """
    Override EmailField to convert emails to lowercase before saving.
    """

    def to_python(self, value):
        """
        Convert email to lowercase.
        """
        value = super(LowercaseEmailField, self).to_python(value)
        # Value can be None so check that it's a string before lowercasing.
        if isinstance(value, str):
            return value.lower()
        return value


class UserManager(BaseUserManager):
    def create_user(self, email, password=None):
        """
        Creates and saves a User with the given email, date of
        birth and password.
        """
        if not email:
            raise ValueError("Users must have an email address")

        user = self.model(email=self.normalize_email(email), name=email)

        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_superuser(self, email, password):
        """
        Creates and saves a superuser with the given email, date of
        birth and password.
        """
        user = self.create_user(
            email,
            password=password,
        )
        user.is_superuser = True
        user.save(using=self._db)
        org = Organization.objects.filter(name="superusers")
        if org.exists():
            org = org.first()
        else:
            org = Organization(name="superusers")
            org.save()
        org.add_admin(user)
        return user


class User(AbstractBaseUser):
    name = models.CharField(max_length=128, null=True)
    email = LowercaseEmailField(unique=True)
    USERNAME_FIELD = "email"
    objects = UserManager()
    current_organization_id = models.IntegerField(null=True)
    is_superuser = models.BooleanField(default=False)
    notifications = models.OneToOneField(
        Notification, on_delete=models.CASCADE, null=True
    )

    def __str__(self):
        return self.email

    @property
    def is_staff(self):
        "Is the user a member of staff?"
        # Simplest possible answer: All admins are staff
        return self.is_superuser

    def has_perm(self, perm, obj=None):
        return self.is_superuser

    def has_module_perms(self, app_label):
        return self.is_superuser

    def get_current_organization(self):
        return self.organization_set.filter(pk=self.current_organization_id).first()

    @is_staff.setter
    def is_staff(self, value):
        self._is_staff = value


class Organization(models.Model):
    name = models.CharField(max_length=128)
    user = models.ManyToManyField(User)
    admin = models.ManyToManyField(User, related_name="admin_user")

    def __str__(self):
        return self.name

    def add_admin(self, user):
        self.user.add(user)
        self.admin.add(user)


class Invitation(models.Model):
    def __eq__(self, other):
        return self.email == other.email and self.organization == other.organization

    def __hash__(self):
        return hash((self.email, self.organization))

    email = LowercaseEmailField()
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)
    invited_by = models.ForeignKey(User, on_delete=models.CASCADE)
    token = models.CharField(max_length=128)
    datetime = models.DateTimeField(auto_now=True)
    expires_at = models.DateTimeField()
    admin = models.BooleanField(default=False)
    invite_link = models.TextField(null=True)

    def __str__(self):
        return "{0}_invite_to_{1}".format(self.email, self.organization)


class ForgottenPassword(models.Model):
    email = LowercaseEmailField()
    token = models.CharField(max_length=128)
    datetime = models.DateTimeField(auto_now=True)
    expires_at = models.DateTimeField()

    def __str__(self):
        return "{0}_forgotten_password".format(self.email)



import logging

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Workflow

logger = logging.getLogger(__name__)


class TestAPIRegistration(APITestCase):
    def test_registration(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": "foo@bar.com",
                "password": "foowordbar",
                "name": "foo",
                "organization": "barinc",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse("password" in response.data)

    def test_registration_without_organization(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": "foo@bar.com",
                "password": "foowordbar",
                "name": "foo",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_registration_short_password(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": "foo@bar.com",
                "password": "fooword",
                "name": "foo",
                "organization": "barinc",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class TestAPIRegistrationTemplate(APITestCase):
    def setUp(self):
        user_name = "foo"
        user_email = "foo@bar.com"
        template_org_name = "fooinc"
        user_password = "foowordbar"

        user = User(name=user_name, email=user_email)
        user.set_password(user_password)
        user.save()
        self.user_id = user.id
        organization = Organization(name=template_org_name, pk=40)
        organization.save()
        self.org_id = organization.id
        organization.add_admin(user)
        wf_data = [
            {
                "id": "foo",
                "name": "foo",
                "type": "text",
                "text": {"read_only": True},
            },
            {
                "id": "foo",
                "name": "foo",
                "type": "single_selection",
                "single_selection": {"options": ["foo1", "bar1"]},
            },
        ]
        template_workflow = Workflow(
            data=wf_data, name="template1", created_by=user, organization=organization
        )
        template_workflow.save()
        self.template_id = template_workflow.pk

    def test_template_registration(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": "bar@bar.com",
                "password": "foowordbar",
                "name": "bar",
                "organization": "barinc",
                "template_id": self.template_id,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        template_workflow = Workflow.objects.get(pk=self.template_id)
        workflow = Workflow.objects.filter(organization__name="barinc").first()
        self.assertEqual(workflow.data, template_workflow.data)
        self.assertEqual(workflow.name, template_workflow.name)
        self.assertNotEqual(workflow.organization, template_workflow.organization)
        self.assertNotEqual(workflow.created_by, template_workflow.created_by)



# Generated by Django 2.2.10 on 2020-06-26 17:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0012_auto_20200624_1449"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="notification",
            name="last_notified",
        ),
    ]



import functools
from enum import Enum

from django.conf import settings
from google.cloud import storage

from human_lambdas.workflow_handler.latency import timer


class Region(Enum):
    bucket_name: str

    @functools.lru_cache
    def get_bucket(self) -> storage.Bucket:

        if self.bucket_name is None:
            raise ValueError("Regional bucket name is None. Config is missing.")

        with timer(f"Connecting to cloud storage bucket {self.bucket_name}"):
            client = storage.Client()
            return client.get_bucket(self.bucket_name)

    def __new__(cls, bucket_name: str):
        obj = object.__new__(cls)
        obj.bucket_name = bucket_name
        return obj

    AU = settings.REGIONAL_BUCKET_AU
    US = settings.REGIONAL_BUCKET_US



import logging

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User

logger = logging.getLogger(__name__)


class TestOrganizations(APITestCase):
    def setUp(self):
        self.preset_user_name = "foo"
        self.preset_user_email = "foo@bar.com"
        self.organization_name = "fooinc"
        self.preset_user_password = "foowordbar"

        user = User(name=self.preset_user_name, email=self.preset_user_email)
        user.set_password(self.preset_user_password)
        user.save()
        organization = Organization(name=self.organization_name)
        organization.save()
        self.org_id = organization.id
        organization.add_admin(user)
        user = User(name="wrong_username", email="wrong@bar.com")
        user.set_password("wrong_user")
        user.save()

    def test_create_organization(self):
        response = self.client.post(
            "/v1/users/token",
            {"email": "foo@bar.com", "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        data = {"name": "new_org"}
        response = self.client.post("/v1/orgs/create", data, format="json")
        self.assertEqual(response.data["name"], data["name"])
        org_obj = Organization.objects.get(name=data["name"])
        self.assertTrue(org_obj.admin.filter(email="foo@bar.com").exists())

    def test_delete_organization(self):
        response = self.client.post(
            "/v1/users/token",
            {"email": "foo@bar.com", "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.delete("/v1/orgs/%s" % self.org_id)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        user = User.objects.get(email=self.preset_user_email)
        self.assertIsNone(user.current_organization_id)

    def test_get_organization(self):
        response = self.client.post(
            "/v1/users/token",
            {"email": "foo@bar.com", "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.get("/v1/orgs/%s" % self.org_id)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["name"], self.organization_name)

    def test_update_organization(self):
        response = self.client.post(
            "/v1/users/token",
            {"email": "foo@bar.com", "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.patch(
            "/v1/orgs/%s" % self.org_id, {"name": "new name"}, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["name"], "new name")

    def test_list_organization(self):
        response = self.client.post(
            "/v1/users/token",
            {"email": self.preset_user_email, "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.get("/v1/orgs")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], self.organization_name)
        self.assertIsNotNone(response.data[0].get("id"))

    def test_get_organization_wrong_user(self):
        response = self.client.post(
            "/v1/users/token", {"email": "wrong@bar.com", "password": "wrong_user"}
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.get("/v1/orgs/%s" % self.org_id)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_organization_data_deletion(self):
        organization = Organization.objects.all()
        organization.delete()
        self.assertEqual(len(Organization.objects.all()), 0)



# Generated by Django 2.2.10 on 2020-08-19 17:31

from django.db import migrations

import human_lambdas.workflow_handler.fields


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0020_workflownotification_last_notified"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="data",
            field=human_lambdas.workflow_handler.fields.JSONField(
                blank=True, default=[]
            ),
        ),
        migrations.AddField(
            model_name="workflow",
            name="data",
            field=human_lambdas.workflow_handler.fields.JSONField(
                blank=True, default=[]
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2020-02-10 15:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0011_auto_20200129_1640"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="n_tasks",
            field=models.IntegerField(default=0),
        ),
    ]



import factory


class BlockFactory(factory.Factory):
    class Meta:
        abstract = True

    _id = "6c9d6441-5be5-4a15-814c-1c9413e8dd22"
    id = "email_id"
    layout = factory.Dict(
        {
            "h": 2,
            "i": "6c9d6441-5be5-4a15-814c-1c9413e8dd22",
            "minH": 2,
            "minW": 4,
            "moved": False,
            "static": False,
            "w": 5,
            "x": 7,
            "y": 0,
        }
    )
    name = "Email Block Title"


class EmailBlockFactory(BlockFactory):
    email = factory.Dict(
        {
            "placeholder": "adf@asdf.co",
            "read_only": False,
            "use_placeholder": True,
        }
    )
    type = "email"


class EmailFormBlockFactory(BlockFactory):
    form_sequence = factory.Dict(
        {
            "data": factory.List(
                [
                    factory.Dict(
                        {
                            "email": factory.Dict({}),
                            "id": "email_id",
                            "name": "Email Question",
                            "type": "email",
                        }
                    )
                ]
            ),
            "history": [],
            "is_required": True,
        }
    )
    type = "form_sequence"


class DateBlockFactory(BlockFactory):
    date = factory.Dict({"value": None, "placeholder": None})
    type = "date"


def make_block(*args, FACTORY_CLASS=BlockFactory, **kwargs):
    return factory.build(dict, FACTORY_CLASS=FACTORY_CLASS, *args, **kwargs)



import os
from pathlib import Path

import click
import dotenv
from django.core.management.utils import get_random_secret_key

env_path = Path(".human_lambdas/.env")
env_path.parent.mkdir(exist_ok=True)
env_exists = env_path.exists()

if not env_exists:
    env_path.write_text("\n")

dotenv.load_dotenv(env_path)

if not "SECRET_KEY" in os.environ:
    click.echo(
        click.style(
            f"No SECRET_KEY present, storing default key in {env_path.as_posix()}",
            fg="yellow",
        )
    )
    key = get_random_secret_key()
    os.environ["SECRET_KEY"] = key
    dotenv.set_key(env_path, "SECRET_KEY", key)



# Generated by Django 2.2.5 on 2020-02-13 15:29

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0005_invitation"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="user",
            name="is_admin",
        ),
        migrations.AddField(
            model_name="organization",
            name="admin",
            field=models.ManyToManyField(
                related_name="admin_user", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="current_organization_id",
            field=models.IntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="organization",
            name="name",
            field=models.CharField(max_length=128),
        ),
    ]



# Generated by Django 2.2.10 on 2020-08-13 13:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0022_webhook"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="workflow",
            name="hook_id",
        ),
        migrations.AddField(
            model_name="webhook",
            name="Workflow",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.CASCADE,
                to="workflow_handler.Workflow",
            ),
        ),
        migrations.AddField(
            model_name="webhook",
            name="is_zapier",
            field=models.BooleanField(default=False),
        ),
    ]



import os
import shutil
import subprocess
import sys
from pathlib import Path
from subprocess import PIPE

import click


@click.group()
def cli():
    if "POSTGRES_DB" in os.environ:
        click.echo(
            click.style(
                "POSTGRES_DB is set, will try to run against remote DB",
                fg="yellow",
            )
        )
    else:
        click.echo("Running against local sqlite")


@click.command()
def up():
    """Starts Human Lambdas"""
    html = Path(__file__).parent / "html"
    html.mkdir(exist_ok=True)
    html_tgz = Path(__file__).parent / "frontend.tgz"

    shutil.unpack_archive(html_tgz, extract_dir=html)
    click.echo("Human Lambdas web running on http://localhost:8000/")
    gunicorn = Path(sys.executable).parent / "gunicorn"
    cmd = f"{gunicorn.as_posix()} human_lambdas.hl_rest_api.wsgi -b 0.0.0.0:8000 -w 1 -t 1 --timeout 0 --preload"
    click.echo(f"Running {cmd}")
    subprocess.run(
        cmd,
        shell=True,
        stdin=PIPE,
        stdout=PIPE,
        check=True,
    )


@click.command()
def initdb():
    """Initializes the Human Lambdas Database"""
    cmd = f"{sys.executable} -m human_lambdas.manage migrate"
    click.echo(f"Running {cmd}")

    subprocess.run(
        cmd,
        shell=True,
        stdin=PIPE,
        stdout=PIPE,
        check=True,
    )


cli.add_command(up)
cli.add_command(initdb)

if __name__ == "__main__":
    cli()



# Generated by Django 2.2.10 on 2020-08-12 08:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0020_workflownotification_last_notified"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="hook_id",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.DeleteModel(
            name="WorkflowHook",
        ),
    ]



# Generated by Django 2.2.5 on 2020-02-20 11:41

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("workflow_handler", "0012_workflow_n_tasks"),
    ]

    operations = [
        migrations.CreateModel(
            name="WorkflowHook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "event",
                    models.CharField(
                        db_index=True, max_length=64, verbose_name="Event"
                    ),
                ),
                ("target", models.URLField(max_length=255, verbose_name="Target URL")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="workflowhooks",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "workflow",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="workflow_handler.Workflow",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# Generated by Django 2.2.10 on 2020-06-24 14:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0011_auto_20200527_1448"),
    ]

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("enabled", models.BooleanField(default=True)),
                ("last_notified", models.DateTimeField(null=True)),
            ],
        ),
        migrations.AddField(
            model_name="user",
            name="notifications",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="user_handler.Notification",
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2020-02-18 09:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0007_user_is_superuser"),
        ("user_handler", "0006_auto_20200206_1457"),
    ]

    operations = []



# Generated by Django 2.2.10 on 2020-06-26 17:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0019_workflownotification"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflownotification",
            name="last_notified",
            field=models.DateTimeField(null=True),
        ),
    ]



# Generated by Django 2.2.5 on 2020-03-03 09:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0013_workflowhook"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



import uuid

from django.core.management.base import BaseCommand

from human_lambdas.workflow_handler.models import Workflow


class Command(BaseCommand):
    help = "Rename the disabled workflows"

    def handle(self, *args, **options):
        workflows = Workflow.objects.filter(disabled=True).all()
        for workflow in workflows:
            wf_name = workflow.name
            workflow.name = f"{wf_name}-{uuid.uuid4()}"
            workflow.save()
            self.stdout.write(
                self.style.SUCCESS('Successfully renamed workflow "%s"' % workflow.id)
            )



# Generated by Django 2.2.5 on 2020-01-09 16:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0007_auto_20191230_2026"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="disabled",
            field=models.BooleanField(default=False),
        ),
    ]



from rest_framework import serializers

from human_lambdas.user_handler.models import Organization


class OrganizationSerializer(serializers.ModelSerializer):
    class Meta:
        model = Organization
        fields = ["name", "user", "admin", "id"]



import analytics
from django.conf import settings

from .hl_client import enqueue_signup


def alias(new_id, old_id):
    if not settings.DEBUG:
        try:
            analytics.alias(new_id, old_id)
        except:
            pass


def identify(*args):
    if not settings.DEBUG:
        try:
            analytics.identify(*args)
        except:
            pass


def track(*args):
    if not settings.DEBUG:
        try:
            analytics.track(*args)
        except:
            pass


def signup_events(user_obj, organization_obj):
    data = {
        "org_id": organization_obj.pk,
        "org_name": organization_obj.name,
        "user_id": user_obj.pk,
        "email": user_obj.email,
        "name": user_obj.name,
        "is_admin": True,
    }
    identify(
        user_obj.pk,
        data,
    )
    track(user_obj.pk, "New Organization", {"source": "website signup"})
    track(user_obj.pk, "User Signup", {"source": "website"})
    enqueue_signup(user_obj.pk, user_obj.email, user_obj.name)



# Generated by Django 2.2.5 on 2020-01-31 14:41

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0004_auto_20191227_1302"),
    ]

    operations = [
        migrations.CreateModel(
            name="Invitation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("email", models.EmailField(max_length=254)),
                ("token", models.CharField(max_length=128)),
                ("datetime", models.DateTimeField(auto_now=True)),
                ("expires_at", models.DateTimeField()),
                (
                    "invited_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "organization",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="user_handler.Organization",
                    ),
                ),
            ],
        ),
    ]



import copy
import logging
import os

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.tests.constants import (
    ALPHA,
    BETA,
    GAMMA,
    REGISTRATION_DATA,
)

from . import DATA_PATH

logger = logging.getLogger(__name__)


class TestCompletedTasksToCSV(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]

        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        self.workflow_data = {
            "name": "uploader",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": [
                            {"id": "foo2", "name": "foo2"},
                            {"id": "bar2", "name": "bar2"},
                        ],
                    },
                },
                {
                    "id": "secondary_output",
                    "name": "barbar",
                    "type": "single_selection",
                    "single_selection": {
                        "options": [
                            {"id": "foo2", "name": "SO1"},
                            {"id": "bar2", "name": "SO2"},
                        ],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.workflow_data,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_list_completed_task_internal_404(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        data = {
            "queue_id": self.workflow_id,
        }
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/completed-tasks-csv".format(self.org_id),
            data=data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_completed_task_internal(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        for i in range(self.total_rows):
            response = self.client.get(
                "/v1/orgs/{0}/queues/{1}/tasks/next".format(
                    self.org_id, self.workflow_id
                ),
                format="json",
            )

            task = response.data
            update_data = copy.deepcopy(task["data"])
            for idata in update_data:
                if idata["id"] == "foo":
                    if i % 2 == 0:
                        idata["single_selection"]["value"] = "foo2"
                    else:
                        idata["single_selection"]["value"] = "bar2"
                elif idata["id"] == "secondary_output":
                    if i % 2 == 0:
                        idata["single_selection"]["value"] = "SO1"
                    else:
                        idata["single_selection"]["value"] = "SO2"
            response_data = {"data": update_data, "assigned_to": self.user_id}
            _ = self.client.patch(
                "/v1/orgs/{0}/queues/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task["id"]
                ),
                data=response_data,
                format="json",
            )
        data = {
            "queue_id": self.workflow_id,
        }
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/completed-tasks-csv".format(self.org_id),
            data=data,
            format="json",
        )
        needed_csv = repr(
            """Alpha,Beta,Gamma,foo,secondary_output
1,2,3,foo2,SO1
4,5,6,bar2,SO2
7,8,9,foo2,SO1""".strip()
        )
        received_csv = repr(
            response.getvalue().decode("utf-8").rstrip().replace("\r", "")
        )
        self.assertEqual(needed_csv, received_csv)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_completed_task_internal_add_output(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        for i in range(self.total_rows - 1):
            response = self.client.get(
                "/v1/orgs/{0}/queues/{1}/tasks/next".format(
                    self.org_id, self.workflow_id
                ),
                format="json",
            )

            task = response.data
            update_data = copy.deepcopy(task["data"])
            for idata in update_data:
                if idata["id"] == "foo":
                    if i % 2 == 0:
                        idata["single_selection"]["value"] = "foo2"
                    else:
                        idata["single_selection"]["value"] = "bar2"
                elif idata["id"] == "secondary_output":
                    if i % 2 == 0:
                        idata["single_selection"]["value"] = "SO1"
                    else:
                        idata["single_selection"]["value"] = "SO2"
            response_data = {"data": update_data, "assigned_to": self.user_id}

            response = self.client.patch(
                "/v1/orgs/{0}/queues/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task["id"]
                ),
                data=response_data,
                format="json",
            )
        self.workflow_data["data"].append(
            {
                "id": "tertiary_output",
                "name": "terttert",
                "type": "single_selection",
                "single_selection": {
                    "options": [
                        {"id": "fooalpha2", "name": "TO1"},
                        {"id": "barbeta2", "name": "TO2"},
                    ],
                },
            }
        )

        response = self.client.patch(
            "/v1/orgs/{0}/queues/{1}".format(self.org_id, self.workflow_id),
            self.workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

        response = self.client.get(
            "/v1/orgs/{0}/queues/{1}/tasks/next".format(self.org_id, self.workflow_id),
            format="json",
        )

        task = response.data
        update_data = copy.deepcopy(task["data"])
        for idata in update_data:
            if idata["id"] == "foo":
                idata["single_selection"]["value"] = "bar2"
            elif idata["id"] == "secondary_output":
                idata["single_selection"]["value"] = "SO2"
            elif idata["id"] == "tertiary_output":
                idata["single_selection"]["value"] = "TO2"
        response_data = {"data": update_data, "assigned_to": self.user_id}

        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                self.org_id, self.workflow_id, task["id"]
            ),
            data=response_data,
            format="json",
        )
        data = {
            "queue_id": self.workflow_id,
        }
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/completed-tasks-csv".format(self.org_id),
            data=data,
            format="json",
        )
        needed_csv = repr(
            """Alpha,Beta,Gamma,foo,secondary_output,tertiary_output
1,2,3,foo2,SO1,
4,5,6,bar2,SO2,
7,8,9,bar2,SO2,TO2""".strip()
        )
        received_csv = repr(
            response.getvalue().decode("utf-8").rstrip().replace("\r", "")
        )
        self.assertEqual(needed_csv, received_csv)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_completed_task_internal_remove_output(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        for i in range(self.total_rows - 1):
            response = self.client.get(
                "/v1/orgs/{0}/queues/{1}/tasks/next".format(
                    self.org_id, self.workflow_id
                ),
                format="json",
            )

            task = response.data
            update_data = copy.deepcopy(task["data"])
            for idata in update_data:
                if idata["id"] == "foo":
                    if i % 2 == 0:
                        idata["single_selection"]["value"] = "foo2"
                    else:
                        idata["single_selection"]["value"] = "bar2"
                elif idata["id"] == "secondary_output":
                    if i % 2 == 0:
                        idata["single_selection"]["value"] = "SO1"
                    else:
                        idata["single_selection"]["value"] = "SO2"
            response_data = {"data": update_data, "assigned_to": self.user_id}
            response = self.client.patch(
                "/v1/orgs/{0}/queues/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task["id"]
                ),
                data=response_data,
                format="json",
            )

        self.workflow_data["data"] = [
            idata
            for idata in self.workflow_data["data"]
            if idata["id"] != "secondary_output"
        ]

        response = self.client.patch(
            "/v1/orgs/{0}/queues/{1}".format(self.org_id, self.workflow_id),
            self.workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

        response = self.client.get(
            "/v1/orgs/{0}/queues/{1}/tasks/next".format(self.org_id, self.workflow_id),
            format="json",
        )

        task = response.data
        update_data = copy.deepcopy(task["data"])

        for idata in update_data:
            if idata["id"] == "foo":
                idata["single_selection"]["value"] = "bar2"
        response_data = {"data": update_data, "assigned_to": self.user_id}
        response = self.client.patch(
            "/v1/orgs/{0}/queues/{1}/tasks/{2}".format(
                self.org_id, self.workflow_id, task["id"]
            ),
            data=response_data,
            format="json",
        )

        data = {
            "queue_id": self.workflow_id,
        }
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/completed-tasks-csv".format(self.org_id),
            data=data,
            format="json",
        )
        needed_csv = repr(
            """Alpha,Beta,Gamma,foo
1,2,3,foo2
4,5,6,bar2
7,8,9,bar2""".strip()
        )
        received_csv = repr(
            response.getvalue().decode("utf-8").rstrip().replace("\r", "")
        )
        self.assertEqual(needed_csv, received_csv)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_no_workflow_id(self):
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/completed-tasks-csv".format(self.org_id)
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)



from django.core.management.base import BaseCommand

from human_lambdas.workflow_handler.models import Workflow


class Command(BaseCommand):
    help = "Rename the workflows with same names"

    def handle(self, *args, **options):
        workflows = Workflow.objects.all()
        for workflow in workflows:
            wf_name = workflow.name
            n_wf_names = Workflow.objects.filter(name=wf_name).count()
            if n_wf_names > 1:
                workflow.name = f"{wf_name}-{n_wf_names}"
                workflow.save()
                self.stdout.write(
                    self.style.SUCCESS(
                        'Successfully renamed workflow "%s"' % workflow.id
                    )
                )



import logging

from django.test.utils import override_settings
from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import ForgottenPassword, User

logger = logging.getLogger(__name__)


class TestInvite(APITestCase):
    def setUp(self):
        self.valid_token = "thisisavalidtoken"
        user = User(name="test", email="foowordfoo@bar.com")
        user.save()
        forgotten_password = ForgottenPassword(
            email="foowordfoo@bar.com",
            token=self.valid_token,
            expires_at=timezone.now() + timezone.timedelta(15),
        )
        forgotten_password.save()

    @override_settings(DEBUG=True)
    def test_forgotten_password(self):
        data = {"email": "sean@humanlambdas.com"}
        response = self.client.post("/v1/users/forgotten-password", data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @override_settings(DEBUG=True)
    def test_forgotten_password_bad_email(self):
        data = {"email": "aaa.com"}
        response = self.client.post("/v1/users/forgotten-password", data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_forgotten_password_bad_token(self):
        response = self.client.get(
            "/v1/users/forgotten-password-token/feo80w3fn83t4f2n0fnwf3wb793282fsu"
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_forgotten_password_good_token(self):
        response = self.client.get(
            "/v1/users/forgotten-password-token/{0}".format(self.valid_token)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    # post endpoint after this line
    def test_forgotten_password_post_wrong_token(self):
        response = self.client.post(
            "/v1/users/forgotten-password-token/{0}".format("thisisnotavalidtoken"),
            {"password": "longlong"},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_forgotten_password_post_no_password(self):
        response = self.client.post(
            "/v1/users/forgotten-password-token/{0}".format(self.valid_token),
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_forgotten_password_post_short_password(self):
        response = self.client.post(
            "/v1/users/forgotten-password-token/{0}".format(self.valid_token),
            {"password": "short"},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_forgotten_password_post_expired_token(self):
        token = "thisisaanothersampletoken"
        forgotten_password = ForgottenPassword(
            email="foowordfoo@bar.com",
            token=token,
            expires_at=timezone.now() - timezone.timedelta(15),
        )
        forgotten_password.save()
        response = self.client.post(
            "/v1/users/forgotten-password-token/{0}".format(token),
            {"password": "feefeefee"},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_forgotten_password_post(self):
        token, email, password, new_password = (
            "thisisthecorrectsampletoken",
            "foo@bar.com",
            "foofoofoo",
            "feefeefee",
        )
        user = User(name="sample", email=email, password=password)
        user.save()
        forgotten_password = ForgottenPassword(
            email=email,
            token=token,
            expires_at=timezone.now() + timezone.timedelta(15),
        )
        forgotten_password.save()
        response = self.client.post(
            "/v1/users/forgotten-password-token/{0}".format(token),
            {"password": new_password},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        response = self.client.post(
            "/v1/users/token", {"email": email, "password": new_password}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)



from django.urls import include, path

from human_lambdas.organization_handler.views import (
    CreateOrganization,
    GetOrganizationView,
    ListOrganizationView,
    ListOrgUsersView,
    RetrieveUpdateRemoveUserOrgView,
)
from human_lambdas.user_handler.invitation import SendInviteView

urlpatterns = [
    # organization
    path("", ListOrganizationView.as_view(), name="list-organization"),
    path("/create", CreateOrganization.as_view(), name="create-organization"),
    path("/<int:org_id>", GetOrganizationView.as_view(), name="get-organization"),
    # organization user management
    path(
        "/<int:org_id>/users",
        ListOrgUsersView.as_view(),
        name="list-organization-users",
    ),
    path(
        "/<int:org_id>/users/<int:user_id>",
        RetrieveUpdateRemoveUserOrgView.as_view(),
        name="remove-organization-users",
    ),
    # organization invite
    path("/<int:org_id>/invite", SendInviteView.as_view(), name="send-invite"),
    # workflows
    path("/<int:org_id>/workflows", include("human_lambdas.v1.workflows")),
    path("/<int:org_id>/queues", include("human_lambdas.v1.workflows")),
    #  metrics
    path("/<int:org_id>/metrics", include("human_lambdas.v1.metrics"), name="metrics"),
]



from django.http.request import HttpRequest
from django.utils.deprecation import MiddlewareMixin

# WARNING: These must be kept up to date with all path='.*' in the client, and cannot clash
# with external.urls
CLIENT_ONLY = [
    "invite",
    "forgot",
    "signup",
    "signout",
    "outstanding",
    "queues",
    "users",
    "audits",
    "metrics",
    "settings",
]


def is_client_only(path: str):
    return any([path.startswith(f"/{cc}") for cc in CLIENT_ONLY])


class RewriteMiddleware(MiddlewareMixin):
    """Provide rewrites so client side web routes work, e.g. deeplinking into /queues"""

    def process_request(self, request: HttpRequest):
        if is_client_only(request.path_info):
            request.path_info = "/"
        pass



from unittest.mock import patch

import jwt
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.hl_rest_api.settings import (
    GOOGLE_OAUTH_CLIENT_ID,
    SECRET_KEY,
)
from human_lambdas.user_handler.models import User


class TestGoogleAuth(APITestCase):
    email = "foo@bar.com"
    name = "foo"
    organization = "barinc"
    mock_token = "blah"

    def mock_verify(self, token, request, client_id):
        if token != self.mock_token:
            raise Exception()
        if client_id != GOOGLE_OAUTH_CLIENT_ID:
            raise Exception()
        return {"email": "foo@bar.com"}

    def isUserAuthed(self, id):
        response = self.client.get(f"/v1/users/{id}")
        return response.status_code == status.HTTP_200_OK

    def test_when_register_without_password_then_ok(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": self.email,
                "name": self.name,
                "organization": self.organization,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        assert User.objects.get(**response.data)

    def test_when_google_token_passed_then_jwt_returned(self):
        google_user = User.objects.create(
            email=self.email, name=self.name, organization=self.organization
        )

        with patch("google.oauth2.id_token.verify_oauth2_token") as verify:
            verify.side_effect = self.mock_verify

            response = self.client.post(
                "/v1/users/token/google",
                {
                    "token": self.mock_token,
                },
            )
            access_token = response.data["access"]

            decoded = jwt.decode(access_token, SECRET_KEY)
            assert decoded["user_id"] == google_user.id
            assert not self.isUserAuthed(google_user.id)
            self.client.credentials(HTTP_AUTHORIZATION=f"Bearer {access_token}")
            assert self.isUserAuthed(google_user.id)

    def test_when_password_signin_on_google_account_then_fails(self):
        User.objects.create(
            email=self.email, name=self.name, organization=self.organization
        )

        response = self.client.post(
            "/v1/users/token",
            {
                "email": self.email,
                "password": "a",
            },
        )
        pass
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(
            response.data["errors"][0]["message"],
            "This account only uses Google sign in",
        )

    def test_when_no_user_for_token_then_bad_request(self):
        with patch("google.oauth2.id_token.verify_oauth2_token") as verify:
            verify.side_effect = self.mock_verify

            response = self.client.post(
                "/v1/users/token/google",
                {
                    "token": self.mock_token,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_when_no_token_then_bad_request(self):
        response = self.client.post(
            "/v1/users/token/google",
            {},
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_when_token_is_malformed_then_bad_request(self):
        response = self.client.post(
            "/v1/users/token/google",
            {
                "token": "iuhiohoiuh",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User


class TestUserRoles(APITestCase):
    def setUp(self):
        self.admin_email = "foo@bar.com"
        user = User(name="foo", email=self.admin_email)
        user.set_password("foopwdbar")
        user.save()
        self.organization_name = "fooinc"
        organization = Organization(name=self.organization_name)
        organization.save()
        organization.add_admin(user)
        self.org_id = organization.pk
        user.current_organization_id = self.org_id
        user.save()
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foopwdbar"}
        )
        self.admin_access_token = response.data["access"]
        self.admin_id = user.pk

        self.worker_email = "worker@bar.com"
        worker = User(
            name="peter", email=self.worker_email, current_organization_id=self.org_id
        )
        worker.set_password("pwdbarbar")
        worker.save()
        organization.user.add(worker)
        response = self.client.post(
            "/v1/users/token", {"email": "worker@bar.com", "password": "pwdbarbar"}
        )
        self.worker_access_token = response.data["access"]
        self.worker_id = worker.pk

    def test_admin_list_user(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.admin_access_token)
        response = self.client.get("/v1/orgs/{0}/users".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 2)
        self.assertTrue(self.admin_email in [i["email"] for i in response.data])
        self.assertTrue(self.worker_email in [i["email"] for i in response.data])

    def test_worker_list_user(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.worker_access_token)
        response = self.client.get("/v1/orgs/{}/users".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 2)

    def test_admin_retrieve(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.admin_access_token)
        response = self.client.get(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.admin_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(self.admin_email, response.data["email"], response.data)
        response = self.client.get(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.worker_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(self.worker_email, response.data["email"], response.data)

    def test_worker_retrieve(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.worker_access_token)
        response = self.client.get(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.admin_id)
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response.data)
        response = self.client.get(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.worker_id)
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response.data)

    def test_worker_update(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.worker_access_token)
        data = {"admin": True}
        response = self.client.patch(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.worker_id), data
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response.data)
        data = {"admin": False}
        response = self.client.patch(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.admin_id), data
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response.data)

    def test_admin_update(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.admin_access_token)
        data = {"admin": False}
        response = self.client.patch(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.admin_id), data
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN, response.data)
        data = {"admin": True}
        response = self.client.patch(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.worker_id), data
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(response.data["message"], "Users role was changed to admin")

    def test_admin_delete(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.admin_access_token)
        response = self.client.delete(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.worker_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        response = self.client.delete(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.admin_id)
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_worker_delete(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.worker_access_token)
        response = self.client.delete(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.worker_id)
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        response = self.client.delete(
            "/v1/orgs/{0}/users/{1}".format(self.org_id, self.admin_id)
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



# Generated by Django 2.2.10 on 2020-05-04 23:34

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("workflow_handler", "0015_auto_20200309_1733"),
    ]

    operations = [
        migrations.AlterField(
            model_name="workflow",
            name="name",
            field=models.CharField(max_length=128, unique=True),
        ),
        migrations.CreateModel(
            name="Source",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="workflow_handler.Workflow",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="task",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="workflow_handler.Source",
            ),
        ),
    ]



# Generated by Django 2.2.13 on 2020-11-18 11:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0031_task_correct"),
    ]

    operations = [
        migrations.AlterField(
            model_name="taskactivity",
            name="action",
            field=models.CharField(
                choices=[
                    ("created", "created"),
                    ("assigned", "assigned"),
                    ("completed", "completed"),
                    ("saved", "saved"),
                    ("comment", "comment"),
                    ("audited_correct", "audited_correct"),
                    ("audited_incorrect", "audited_incorrect"),
                    ("audited_empty", "audited_empty"),
                ],
                max_length=128,
            ),
        ),
    ]



# Generated by Django 2.2.13 on 2020-11-03 17:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0029_auto_20201029_1449"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="handling_time_seconds",
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name="task",
            name="session_started_at",
            field=models.DateTimeField(null=True),
        ),
    ]



from rest_framework import status

from human_lambdas.user_handler.models import Notification, Organization, User
from human_lambdas.workflow_handler.models import Workflow
from human_lambdas.workflow_handler.tests.constants import (
    GUIDELINES_URL,
    INTERNAL_WORKER_REGISTRATION_DATA,
    REGISTRATION_DATA,
    SUPER_ADMIN_REGISTRATION_DATA,
    TASK_DESCRIPTION,
    WORKFLOW_DATA,
)
from human_lambdas.workflow_handler.tests.util import HLTestCase
from human_lambdas.workflow_handler.utils import STAFF_ORG_ID, TEMPLATE_ORG_ID


class TestCRUDWorkflow(HLTestCase):
    def setUp(self):

        self.access_token_super_admin = self.register(
            SUPER_ADMIN_REGISTRATION_DATA, is_super_admin=True
        )["access"]
        self.access_token_internal_worker = self.register(
            INTERNAL_WORKER_REGISTRATION_DATA, is_internal_worker=True
        )["access"]

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.user_org_id = Organization.objects.get(user__email="foo@bar.com").pk
        self.org_id = self.user_org_id
        self.access_token = response.data["access"]

        # perhaps need an internal endpoint to make user as a worker of an org
        org = Organization.objects.get(pk=self.org_id)
        notification = Notification()
        notification.save()
        user = User(name="worker", email="worker@bar.com", notifications=notification)
        user.set_password("foowordbar")
        user.save()
        org.user.add(user)
        self.user = user
        response = self.client.post(
            "/v1/users/token", {"email": "worker@bar.com", "password": "foowordbar"}
        )
        self.access_token_worker = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

    def test_create_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.assertTrue(Workflow.objects.filter(name=workflow_data["name"]).exists())

    def test_create_workflow_worker(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token_worker)

        workflow_data = {
            "name": "workerwf",
            "data": [
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "text",
                    "layout": {},
                    "text": {"read_only": True},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        # HOTFIX: temporarily remove
        # self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN, response.data)
        self.assertFalse(Workflow.objects.filter(name=workflow_data["name"]).exists())

    def test_retrieve_template(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        templates_org = Organization(id=TEMPLATE_ORG_ID, name="")
        templates_org.save()
        workflow_id = 777
        Workflow(
            id=workflow_id,
            name="template1",
            created_by=self.user,
            data={},
            organization=templates_org,
        ).save()

        response = self.client.get(
            f"/v1/orgs/{TEMPLATE_ORG_ID}/workflows/{workflow_id}"
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_retrieve_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow_obj = Workflow.objects.filter(name=workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_retrieve_workflow_worker(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow_obj = Workflow.objects.filter(name=workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token_worker)
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_update_workflow_worker(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow_obj = Workflow.objects.filter(name=workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        updated_text = "not so great wf"
        updated_workflow_data = {
            "name": updated_text,
        }
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token_worker)
        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        # HOTFIX: temporarily remove
        # self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN, response.data)

        workflow_obj = Workflow.objects.filter(name=workflow_data["name"])
        workflow = workflow_obj.first()
        # self.assertNotEqual(updated_text, workflow.name)

    def test_when_running_workflow_then_there_must_be_description(self):
        workflow_data_no_desc = {**WORKFLOW_DATA}

        workflow_id = self.create_workflow(workflow_data_no_desc, org_id=self.org_id)[
            "id"
        ]
        assert Workflow.objects.get(pk=workflow_id).is_running == False

        running_workflow_data_no_desc = {
            **workflow_data_no_desc,
            "is_running": True,
        }

        response = self.client.patch(
            f"/v1/orgs/{self.org_id}/workflows/{workflow_id}",
            running_workflow_data_no_desc,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        assert Workflow.objects.get(pk=workflow_id).is_running == False
        assert Workflow.objects.get(pk=workflow_id).task_description is None

        running_workflow_data = {
            **workflow_data_no_desc,
            "is_running": True,
            "task_description": TASK_DESCRIPTION,
        }
        response = self.client.patch(
            f"/v1/orgs/{self.org_id}/workflows/{workflow_id}",
            running_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        wf = Workflow.objects.get(pk=workflow_id)
        assert wf.is_running == True
        assert wf.task_description == TASK_DESCRIPTION

    def test_when_creating_workflow_then_running_data_is_collected(self):
        workflow_data_running_no_desc = {**WORKFLOW_DATA, "is_running": True}

        response = self.client.post(
            f"/v1/orgs/{self.org_id}/workflows/create",
            workflow_data_running_no_desc,
            format="json",
        )

        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data_running = {
            **workflow_data_running_no_desc,
            "task_description": TASK_DESCRIPTION,
            "guidelines_url": GUIDELINES_URL,
        }
        workflow_id = self.create_workflow(workflow_data_running, org_id=self.org_id)[
            "id"
        ]
        wf = Workflow.objects.get(pk=workflow_id)
        assert wf.is_running == True
        assert wf.task_description == TASK_DESCRIPTION

    def test_when_staff_worker_lists_workflows_then_all_running_returned(self):
        self.set_user(self.access_token, org_id=str(self.org_id))
        workflow_data = {
            **WORKFLOW_DATA,
            "is_running": True,
            "task_description": TASK_DESCRIPTION,
        }
        self.create_workflow(workflow_data)

        self.set_user(self.access_token_internal_worker, org_id=STAFF_ORG_ID)

        data = self.list_workflows(org_id=self.org_id)
        assert data[0]["org_id"] == self.user_org_id

    def test_update_workflow_super_admin(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = WORKFLOW_DATA
        self.create_workflow(workflow_data, org_id=self.org_id)

        workflow_obj = Workflow.objects.filter(name=workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        updated_text = "not so great wf"
        updated_workflow_data = {
            "name": updated_text,
        }
        self.client.credentials(
            HTTP_AUTHORIZATION="Bearer " + self.access_token_super_admin
        )
        get_response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        assert get_response.status_code == status.HTTP_200_OK
        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_update_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow_obj = Workflow.objects.filter(name=workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        assert workflow.pinned_block == None
        workflow_data2 = {"name": "foowf2", "data": WORKFLOW_DATA["data"]}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data2,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)

        updated_text = "not so great wf"
        pinned_block_id = "my_block"
        updated_workflow_data = {"name": updated_text, "pinned_block": pinned_block_id}

        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(updated_text, response.data["name"])
        self.assertEqual(pinned_block_id, response.data["pinned_block"])
        workflow_obj = Workflow.objects.filter(name=response.data["name"])
        workflow = workflow_obj.first()
        self.assertEqual(updated_text, workflow.name)
        updated_workflow_data = {
            "name": workflow_data["name"],
        }

        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        updated_workflow_data = {
            "name": workflow_data2["name"],
        }

        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data1 = {**WORKFLOW_DATA, "name": "foowf"}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data1,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow_data2 = {**WORKFLOW_DATA, "name": "foowf2"}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data2,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id))
        result_1, result_2 = sorted(response.data, key=lambda dd: dd["name"])
        self.assertTrue(result_1.pop("id"))
        self.assertTrue(result_2.pop("id"))
        self.assertEqual(result_1.pop("n_tasks"), 0)
        self.assertEqual(result_2.pop("n_tasks"), 0)
        self.assertTrue(result_1.pop("created_at"))
        self.assertTrue(result_2.pop("created_at"))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        [idata.pop("active_users") for idata in response.data]

        del workflow_data1["data"]
        del workflow_data2["data"]
        del result_1["data"]
        del result_2["data"]

        self.assertEqual(workflow_data1, result_1, response.data)
        self.assertEqual(workflow_data2, result_2, response.data)

    def test_list_workflow2(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data1 = {**WORKFLOW_DATA, "name": "foowf"}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data1,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        workflow_data2 = {**WORKFLOW_DATA, "name": "foowf2"}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data2,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)

        workflow_obj = Workflow.objects.filter(name=workflow_data2["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        updated_workflow_data = {
            "disabled": True,
        }
        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            updated_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id))
        result_1 = response.data[0]
        self.assertTrue(result_1.pop("id"))
        self.assertTrue(result_1.pop("created_at"))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 1)
        workflow_data1["n_tasks"] = 0
        result_1.pop("active_users")
        del workflow_data1["data"]
        del result_1["data"]
        self.assertEqual(workflow_data1, result_1, response.data)
        self.assertNotEqual(Workflow.objects.get(pk=workflow.pk).name, workflow.name)



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Workflow
from human_lambdas.workflow_handler.tests.constants import (
    DELTA,
    REGISTRATION_DATA,
)


class TestErrorPayloadStructure(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]
        self.invalid_token = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
            ".eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0Ij"
            "oxNTE2MjM5MDIyfQ.SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
        )
        self.complete_workflow_data = {
            "name": "foowf",
            "data": [
                DELTA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": ["foo1", "bar1"],
                    },
                },
            ],
        }

    # Create workflow endpoint tests
    def test_create_workflow_no_jwt(self):
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_create_workflow_invalid_jwt(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.invalid_token)
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_create_workflow_missing_data(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        workflow_data = {
            "data": [
                DELTA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": ["foo1", "bar1"],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_workflow_wrong_method(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.patch(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    # test list workflows
    def test_list_workflow_no_jwt(self):
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_workflow_invalid_jwt(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.invalid_token)
        response = self.client.get("/v1/orgs/{}/workflows".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_workflow_wrong_method(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.post("/v1/orgs/{}/workflows".format(self.org_id))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    # test update workflows
    def test_update_workflow_no_jwt(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.content
        )
        self.client.credentials(HTTP_AUTHORIZATION="")
        workflow_obj = Workflow.objects.filter(name=self.complete_workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        self.complete_workflow_data = {
            "name": "not so great wf",
        }
        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_workflow_invalid_jwt(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.content
        )
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.invalid_token)
        workflow_obj = Workflow.objects.filter(name=self.complete_workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        self.complete_workflow_data = {
            "name": "not so great wf",
        }
        response = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_workflow_wrong_method(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.content
        )
        workflow_obj = Workflow.objects.filter(name=self.complete_workflow_data["name"])
        self.assertTrue(workflow_obj.exists())
        workflow = workflow_obj.first()
        self.complete_workflow_data = {
            "name": "not so great wf",
        }
        response = self.client.post(
            "/v1/orgs/{0}/workflows/{1}".format(self.org_id, workflow.pk),
            self.complete_workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)



# Generated by Django 2.2.10 on 2020-06-24 14:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0012_auto_20200624_1449"),
        ("workflow_handler", "0018_auto_20200618_0949"),
    ]

    operations = [
        migrations.CreateModel(
            name="WorkflowNotification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("enabled", models.BooleanField(default=True)),
                (
                    "notification",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="user_handler.Notification",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="workflow_handler.Workflow",
                    ),
                ),
            ],
        ),
    ]



import logging
import uuid

from django.conf import settings
from django.db import transaction
from django.db.models import F, Q
from django.utils import timezone
from rest_framework import exceptions, serializers

from human_lambdas.data_handler.data_validation import (
    DataValidationError,
    data_validation,
)
from human_lambdas.hl_rest_api import analytics
from human_lambdas.hl_rest_api.utils import is_valid_url
from human_lambdas.user_handler.models import Organization

from .models import (
    Source,
    Task,
    TaskActivity,
    WebHook,
    Workflow,
    WorkflowNotification,
)
from .utils import get_session_duration_seconds, notify_staff_run_status

logger = logging.getLogger(__name__)


def clean_form_sequence(data):
    for idata in data:
        if idata["type"] == "form_sequence":
            for form_data in idata["form_sequence"]["data"]:
                form_data[form_data["type"]]["value"] = None


def _validate_automation(instance: Workflow):
    if instance.is_running and instance.task_description is None:
        raise serializers.ValidationError(
            "You must have a task description to automate your queue"
        )


class HookSerializer(serializers.ModelSerializer):
    def validate_event(self, event):
        if event not in settings.HOOK_EVENTS:
            err_msg = "Unexpected event {}".format(event)
            raise exceptions.ValidationError(detail=err_msg)
        return event

    class Meta:
        model = WebHook
        fields = ("target", "user", "event", "workflow", "id")
        read_only_fields = ("user", "event", "workflow", "id")

    def create(self, validated_data):
        validated_data["event"] = "task.completed"
        validated_data["user"] = self.context["request"].user
        validated_data["workflow"] = Workflow.objects.get(
            pk=self.context["view"].kwargs["workflow_id"]
        )
        webhook = WebHook(**validated_data)
        webhook.save()
        analytics.track(
            self.context["request"].user.pk,
            "Register Webhook",
            {"workflow_id": self.context["view"].kwargs["workflow_id"]},
        )
        return webhook


class WorkflowSerializer(serializers.ModelSerializer):
    webhook = HookSerializer(required=False, allow_null=True, write_only=True)
    active_users = serializers.SerializerMethodField(
        "get_active_users", allow_null=True
    )
    org_id = serializers.SerializerMethodField("get_org_id", read_only=True)

    class Meta:
        model = Workflow
        fields = [
            "id",
            "name",
            "disabled",
            "n_tasks",
            "created_at",
            "webhook",
            "active_users",
            "org_id",
            "data",
            "is_running",
            "task_description",
            "guidelines_url",
            "pinned_block",
        ]
        extra_kwargs = {
            "disabled": {"write_only": True},
            "n_tasks": {"read_only": True},
            "id": {"read_only": True},
            "created_at": {"read_only": True},
            "active_users": {"read_only": True},
            "webhook": {"write_only": True},
        }

    def get_org_id(self, instance):
        return instance.organization_id

    def get_active_users(self, instance):
        return list(
            instance.task_set.filter(
                Q(assigned_to__isnull=False), ~Q(status="completed")
            )
            .distinct()
            .values_list("assigned_to__name", flat=True)
        )

    def create(self, validated_data):
        user = self.context["request"].user
        wf_name = validated_data["name"]
        data = validated_data["data"]
        clean_form_sequence(data)
        organization_obj = Organization.objects.filter(
            user=user, pk=int(self.context["view"].kwargs["org_id"])
        )
        if organization_obj.exists() and organization_obj.count() == 1:
            organization = organization_obj.first()
        else:
            raise serializers.ValidationError("Organization not found!")
        workflow = Workflow(
            name=wf_name,
            organization=organization,
            created_by=user,
            data=data,
            is_running=validated_data.get("is_running", False),
            guidelines_url=validated_data.get("guidelines_url"),
            task_description=validated_data.get("task_description"),
        )
        _validate_automation(workflow)
        workflow.save()
        analytics.track(
            user.pk, "Created Workflow", {"name": wf_name, "workflow_id": workflow.pk}
        )
        webhook_data = validated_data.get("webhook")
        if webhook_data:
            webhook_data["event"] = "task.completed"
            webhook_data["user"] = user
            webhook_data["workflow"] = workflow
            WebHook.objects.create(**webhook_data)
        for org_user in organization.user.all():
            wfnotification = WorkflowNotification(
                workflow=workflow, notification=org_user.notifications, enabled=True
            )
            wfnotification.save()

        return workflow

    def update(self, instance, validated_data):
        workflow_name = validated_data.get("name")
        if workflow_name:
            if workflow_name in instance.organization.workflow_set.filter(
                ~Q(pk=instance.pk)
            ).values_list("name", flat=True):
                raise serializers.ValidationError(
                    "Workflow with same name already exists"
                )
            instance.name = workflow_name
        data = validated_data.get("data")
        if data:
            clean_form_sequence(data)
            instance.data = data
        disabled = validated_data.get("disabled")
        if disabled:
            instance.disabled = disabled
            instance.name = f"{instance.name}-{uuid.uuid4()}"
            for wnotification in WorkflowNotification.objects.filter(
                workflow=instance
            ).all():
                wnotification.enabled = False
                wnotification.save()

        if "is_running" in validated_data:
            instance.is_running = validated_data["is_running"]

        if "guidelines_url" in validated_data:
            instance.guidelines_url = validated_data["guidelines_url"]
            if not instance.guidelines_url is None and not is_valid_url(
                instance.guidelines_url
            ):
                raise serializers.ValidationError("guidelines URL is invalid")

        if "task_description" in validated_data:
            instance.task_description = validated_data["task_description"]

        _validate_automation(instance)
        instance.save()

        if "is_running" in validated_data:
            notify_staff_run_status(validated_data, self.context["request"])

        event_name = "Deleted" if disabled else "Updated"
        analytics.track(
            self.context["request"].user.pk,
            "%s Workflow" % event_name,
            {"name": instance.name, "workflow_id": instance.pk},
        )
        webhook_data = validated_data.get("webhook")
        if webhook_data or self.context.get("remove_webhook"):
            hook_instance = WebHook.objects.filter(workflow=instance)
            if hook_instance.exists():
                hook_instance = hook_instance.first()
                if self.context.get("remove_webhook"):
                    hook_instance.delete()
                else:
                    hook_instance.target = webhook_data["target"]
                    hook_instance.save()
            else:
                webhook_data["event"] = "task.completed"
                webhook_data["user"] = self.context["request"].user
                webhook_data["workflow"] = instance
                WebHook.objects.create(**webhook_data)

        return super(WorkflowSerializer, self).update(instance, validated_data)

    def validate_data(self, data):
        if not data:
            raise serializers.ValidationError("Workflow cannot be empty")
        try:
            return data_validation(data, is_workflow=True)
        except DataValidationError as exception_text:
            raise serializers.ValidationError(exception_text)

    def validate_webhook(self, data):
        if data["target"]:
            if not is_valid_url(data["target"]):
                raise serializers.ValidationError("Not a valid URL set for Webhooks")
        return data


class TaskSerializer(serializers.ModelSerializer):
    workflow = serializers.SerializerMethodField()
    workflow_id = serializers.SerializerMethodField()
    n_comments = serializers.SerializerMethodField()

    def validate_event(self, event):
        if event not in settings.HOOK_EVENTS:
            err_msg = "Unexpected event {}".format(event)
            raise exceptions.ValidationError(detail=err_msg)
        return event

    class Meta:
        model = Task
        fields = [
            "id",
            "status",
            "created_at",
            "assigned_to",
            "completed_at",
            "data",
            "workflow",
            "workflow_id",
            "n_comments",
            "correct",
            "region",
        ]

    def create(self, validated_data):
        logger.info(f"taskserializer: create")
        source_name = validated_data.get("source_name")
        if not source_name:
            raise serializers.ValidationError("No source name given!")
        data = validated_data["data"]
        workflow_id = (
            self.context["view"].kwargs["queue_id"]
            if "queue_id" in self.context["view"].kwargs
            else self.context["view"].kwargs["workflow_id"]
        )
        workflow = Workflow.objects.get(id=workflow_id)
        source = Source(
            name=source_name,
            created_by=self.context["request"].user,
            workflow=workflow,
        )
        source.save()
        region = validated_data.get("region")
        task = Task(
            data=data,
            workflow=workflow,
            source=source,
            region=None if region in ["EU", None] else region,
        )
        task.save()
        TaskActivity(
            task=task,
            source=source_name,
            created_by=self.context["request"].user,
            action="created",
        ).save()
        with transaction.atomic():
            workflow.n_tasks = F("n_tasks") + 1
            workflow.save()

        analytics.track(
            self.context["request"].user.pk,
            "Task Created",
            {
                "org_id": workflow.organization.pk,
                "workflow_id": workflow.id,
                "source": source_name,
            },
        )

        return task

    def update(self, instance, validated_data):
        logger.info(f"taskserializer: update")
        user = self.context["request"].user
        if instance.status == "completed" and not validated_data["force"]:
            raise serializers.ValidationError("Cannot change a completed task")
        elif instance.assigned_to == user:
            instance.data = validated_data.get("data", instance.data)

            instance.handling_time_seconds += get_session_duration_seconds(instance)
            instance.session_started_at = None

            if validated_data["submit_task"]:
                instance.status = "completed"
                instance.completed_at = timezone.now()
                instance.save()
                instance.task_completed(user)
                TaskActivity(task=instance, action="completed", created_by=user).save()
                workflow = instance.workflow
                with transaction.atomic():
                    workflow.n_tasks = F("n_tasks") - 1
                    workflow.save()
            else:
                instance.save()
                TaskActivity(task=instance, action="saved", created_by=user).save()
            event_name = "Completed" if validated_data["submit_task"] else "Saved"
            analytics.track(
                user.pk,
                "Task %s" % event_name,
                {
                    "org_id": instance.workflow.organization.pk,
                    "workflow_id": instance.workflow.id,
                    "source": instance.source.name,
                },
            )
        else:
            raise serializers.ValidationError("User not assigned to task")
        return instance

    def validate_data(self, data):
        try:
            return data_validation(data)
        except DataValidationError as exception_text:
            raise serializers.ValidationError(exception_text)

    def get_workflow(self, obj):
        return obj.workflow.name

    def get_workflow_id(self, obj):
        return obj.workflow.pk

    def get_n_comments(self, obj):
        return obj.taskactivity_set.filter(action="comment").count()

    def to_representation(self, instance):
        return instance.get_formatted_task()


class TaskMetadataSerializer(TaskSerializer):
    def to_representation(self, instance: Task):
        return instance.get_formatted_task(include_data=False)


class PendingTaskSerializer(TaskMetadataSerializer):
    def to_representation(self, instance):
        return instance.get_updated_status()


class SourceSerializer(serializers.ModelSerializer):
    class Meta:
        fields = "__all__"
        model = Source



# Generated by Django 2.2.13 on 2021-02-09 15:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0032_auto_20201118_1109"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="workflow",
            name="description",
        ),
    ]



from django.core.management.base import BaseCommand
from django.db.models import Q

from human_lambdas.workflow_handler.models import Task, Workflow


class Command(BaseCommand):
    help = "Updates the task count"

    def handle(self, *args, **options):
        workflows = Workflow.objects.all()
        for workflow in workflows:
            workflow.n_tasks = (
                Task.objects.defer("data")
                .filter(workflow=workflow)
                .filter(~Q(status="completed"))
                .count()
            )
            workflow.save()
            self.stdout.write(
                self.style.SUCCESS(
                    'Successfully updated task count for workflow "%s"' % workflow.id
                )
            )
